# -*- coding: utf-8 -*-
"""auxiliary routines for unit testing"""

# stdlib modules
from json import dumps, load, loads, JSONEncoder
from inspect import currentframe, getouterframes
from pathlib import Path
from difflib import Differ

from casadi import SX

SX.__json__ = SX.__str__
FILENAME = "refdata.json"


class CustomEncoder(JSONEncoder):
    """Custom encoder for Objects"""

    def default(self, o):
        """Try to json-encode objects by calling their ``__json__`` method"""
        try:
            return o.__json__()
        except AttributeError:
            return super().default(o)


def user_agree(message):
    """Ask for confirmation, but assume "no", if not run interactively"""
    try:
        ans = input(f"{message} y/[n]? ")
        return ans.lower() == "y"
    except (OSError, EOFError):  # run automatically with std streams caught
        return False


def assert_reproduction(result, suffix=None):
    """Assert the json-dump of the data to be the same as before.
    This method will (if run interactively) ask the user to accept the
    new or changed data, if no old reference data exists or if it differs.

    The ``sorted=True`` flag is set when dumping the data, allowing
    (nested) dictionaries to be compared correctly.
    """

    def load_file():
        """try to open refdata file. If it doesn't exist, dump and return
        an empty dictionary"""
        try:
            with open(filename, "r", encoding="utf-8") as file:
                data = load(file)
        except FileNotFoundError:
            data = {}
            with open(filename, "w", encoding="utf-8") as file:
                file.write(dumps(data))
        return data

    frame = getouterframes(currentframe())[1]
    caller_file = Path(frame.filename)
    filename = caller_file.absolute().parent / FILENAME
    ref_data_all = load_file()

    # to align and assure compatibility
    result = loads(dumps(result, cls=CustomEncoder))
    func_name = frame.function  # get name of calling function
    func_name = f"{caller_file.name}::{func_name}"
    if suffix:
        func_name = f"{func_name}_{suffix}"
    ref_data = ref_data_all.get(func_name, None)

    def save_data(data):
        """Save the reference data to the file"""
        ref_data_all[func_name] = data
        with open(filename, "w", encoding="utf-8") as file:
            file.write(dumps(ref_data_all, sort_keys=True, indent=2))

    if ref_data is None:
        msg = (f"No reference data exists for {func_name}. " +
               "The following data is generated now:\n\n" +
               dumps(result, indent=2, sort_keys=True) +
               "\n\nDo you accept this data")
        if user_agree(msg):
            save_data(result)
        else:
            raise AssertionError("Reference data rejected by user")
    else:
        ref = dumps(ref_data, indent=2, sort_keys=True)
        val = dumps(result, indent=2, sort_keys=True)
        try:
            assert ref == val
        except AssertionError:
            differ = Differ()
            diff = differ.compare(ref.splitlines(), val.splitlines())
            msg = (
                f"Deviation from reference data detected ({func_name}):\n"
                + "\n".join(diff) + "\n\nDo you accept the new data")
            if user_agree(msg):
                save_data(result)
            else:
                raise



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module containing constants as quantities.
"""

from math import pi as _pi
from .quantity import Quantity as Qty

PI = Qty(_pi)  # the famous pi
R_GAS = Qty("8.31446261815324 J/(mol*K)")  # gas constant
V_LIGHT = Qty("299792458.0 m/s")  # speed of light
H_PLANCK = Qty("6.62607015e-34 J*s")  # Planck constant
GAMMA_G = Qty("6.67430e-11 m**3/(kg*s**2)")  # Newtonian gravitational constant
N_A = Qty("6.02214076e23  1/mol")  # Avogadro constant
ALPHA = Qty(7.2973525693e-3)  # Fine structure constant
SIGMA = Qty("5.670374419e-8 W/(m**2*K**4)")  # Stefan-Boltzmann constant
F = Qty("96485.33212 C/mol")  # Faraday constant
EPS_0 = Qty("8.8541878128e-12 F/m")  # Vacuum electric permittivity
MU_0 = Qty("1.25663706212 N/A**2")  # Vacuum magnetic permittivity
E_0 = Qty("1.602176634e-19 C")  # Elementary charge
K_B = Qty("1.380649e-23 J/K")  # Boltzmann constant

STD_GRAVITY = Qty("9.80665 m/s**2")  # standard acceleration of gravity



"""This module contains data structures that build on the quantity datatype"""

# stdlibs
from typing import Union, TypeVar, Self
from collections.abc import Callable, Iterable, Mapping

# external libs
import casadi as cas
from pint.errors import DimensionalityError

# internal libs
from . import base_unit, Quantity, SymbolQuantity, qvertcat, qpow, qsqrt
from .types import NestedMap


class ParameterDictionary(dict):
    """This class is a nested dictionary of SymbolQuantities to represent
    parameters with functionality to be populated using the ``register_*``
    methods.
    """

    class SparseMatrix(dict):
        """This helper class represents a nested dictionary that contains
        two levels of keys and values representing a quantity."""

        def pair_items(self):
            """Return an iterator yielding a scalar quantity with the key pair
            for each element in the sub-structure. The elements have the
            shape ``(key_1, key_2, quantity)``."""
            for key_1, second in self.items():
                for key_2, quantity in second.items():
                    yield key_1, key_2, quantity

    def register_scalar(self, key: str, unit: str):
        """Create a scalar quantity and add the structure to the dictionary.
        The given unit is converted to base units before being applied. Calling
        the method returns the created quantity

            >>> pdict = ParameterDictionary()
            >>> print(pdict.register_scalar("speed", "cm/h"))
            speed meter / second

        In this output, ``speed`` is the name of the ``casadi.SX`` node
        representing the magnitude of returned Quantity. The dictionary then
        contains the following entry:

            >>> print(pdict)
            {'speed': <Quantity(speed, 'meter / second')>}
        """
        unit = base_unit(unit)
        quantity = SymbolQuantity(key, unit)
        self[key] = quantity
        return quantity

    def register_vector(self, key: str, sub_keys: Iterable[str],
                        unit: str) -> Quantity:
        """Create a quantity vector with symbols and add the structure to
        the dictionary. The given unit is converted to base units before being
        applied. Calling the method returns the created quantity

            >>> pdict = ParameterDictionary()
            >>> print(pdict.register_vector("velocity", "xyz", "knot"))
            [velocity.x, velocity.y, velocity.z] meter / second

        The dictionary then contains the following entries:

            >>> from pprint import pprint
            >>> pprint(pdict)
            {'velocity': {'x': <Quantity(velocity.x, 'meter / second')>,
                          'y': <Quantity(velocity.y, 'meter / second')>,
                          'z': <Quantity(velocity.z, 'meter / second')>}}
        """
        unit = base_unit(unit)
        self[key] = {s: SymbolQuantity(f"{key}.{s}", unit) for s in sub_keys}
        return qvertcat(*self[key].values())

    def register_sparse_matrix(self, key: str, pairs: Iterable[tuple[str, str]],
                               unit: str) -> NestedMap[Quantity]:
        """Create a sparse matrix quantity and add the structure to the
        dictionary. The given unit is converted to base units before being
        applied.

            >>> pdict = ParameterDictionary()
            >>> binaries = [("H2O", "CO2"), ("H2O", "CH4")]
            >>> from pprint import pprint
            >>> pprint(pdict.register_sparse_matrix("K_ij", binaries, "K"))
            {'H2O': {'CH4': <Quantity(K_ij.H2O.CH4, 'kelvin')>,
                     'CO2': <Quantity(K_ij.H2O.CO2, 'kelvin')>}}

        After above call, the dictionary contains the following entries:

            >>> from pprint import pprint
            >>> pprint(pdict)
            {'K_ij': {'H2O': {'CH4': <Quantity(K_ij.H2O.CH4, 'kelvin')>,
                              'CO2': <Quantity(K_ij.H2O.CO2, 'kelvin')>}}}

        """
        unit = base_unit(unit)
        res = ParameterDictionary.SparseMatrix({f: {} for f, _ in pairs})
        for first, second in pairs:
            quantity = SymbolQuantity(f"{key}.{first}.{second}", unit)
            res[first][second] = quantity
        self[key] = res
        return res

    def get_quantity(self, *keys):
        """Extract a quantity from the given sequence of key. Being a nested
        dictionary, each key from the argument list is used to navigate into
        the structure. The value of the most inner addressed key is returned.
        For normal usage, this should be of type ``Quantity``."""
        entry = self
        for key in keys:
            entry = entry[key]
        return entry

    def get_vector_quantity(self, *keys):
        """Extract a vector quantity from the given sequence of keys. The
        method extracts the values of the structure below the sequence of
        argument keys, and concatenates them as a single vector property.
        """
        entry = self
        for key in keys:
            entry = entry[key]
        return qvertcat(*entry.values())


_OType = Union[float, Quantity, Mapping[str, Quantity]]
_SType = Union[float, cas.SX]


class QuantityDict(dict[str, Quantity]):
    """Many properties on process modelling level are vectorial. This includes
    any species-specific properties, such as for instance mole fractions,
    chemical potentials or partial enthalpy. By keeping such data in instances
    of this class, they can always be accessed as a dictionary, using the
    bracket-operator (``__get_item__``).

    Additionally, this class supports most arithmetic operations, such as
    ``+, -, *, /, **`` - all of them interpreted element-wise. As two instances
    can have deviating keys (mostly species), there are some rules:

    - missing elements are assumed as zero, and structural zeros are
      omitted, i.e.

      >>> a = QuantityDict({
      ...         "A": Quantity("1 m"),
      ...         "B": Quantity("50 cm")})
      >>> b = QuantityDict({
      ...         "B": Quantity("1 m"),
      ...         "C": Quantity("50 cm")})
      >>> y = a + b
      >>> for key, value in y.items(): print(f"{key}: {value:~}")
      A: 1 m
      B: 150.0 cm
      C: 50 cm

      >>> y = a * b
      >>> for key, value in y.items(): print(f"{key}: {value:~}")
      B: 50 cm * m

    - A missing denominator element in division directly raises
      ``ZeroDivisionError``

      >>> y = a / b
      Traceback (most recent call last):
      ...
      ZeroDivisionError: Missing denominator element in QuantityDict division

    - Operations can be mixed with scalar Quantities

      >>> y = a["A"] * b
      >>> for key, value in y.items(): print(f"{key}: {value:~}")
      B: 1 m ** 2
      C: 50 cm * m

    - floats as second operands in binary operators act as dimensionless
      quantities

      >>> y = 3 * a
      >>> for key, value in y.items(): print(f"{key}: {value:~}")
      A: 3 m
      B: 150 cm

      >>> y = 3 + a
      Traceback (most recent call last):
      ...
      pint.errors.DimensionalityError: ...
    """
    @classmethod
    def from_vector_quantity(cls, quantity: Quantity,
                             keys: list[str]) -> Self:
        """As the magnitude of a ``Quantity`` can be a container itself,
        this convenience factory method combines such a vector quantity with
        a set of given keys into a ``QuantityDict`` object

        >>> raw = Quantity([1, 2, 3], "m")
        >>> dic = QuantityDict.from_vector_quantity(raw, ["A", "B", "C"])
        >>> for name, value in dic.items():
        ...     print(f"{name}: {value:~}")
        A: 1 m
        B: 2 m
        C: 3 m
        """
        try:
            l_magnitude = len(quantity.magnitude)
        except TypeError:
            l_magnitude = quantity.magnitude.size1()
        if l_magnitude != len(keys):
            raise ValueError("Dimension mismatch in resolving vector quantity: "
                             f"{l_magnitude} != {len(keys)}")
        return cls({key: quantity[k] for k, key in enumerate(keys)})

    def __add__(self, other: _OType) -> Self:
        try:
            items = other.items()
        except AttributeError:
            return QuantityDict({k: v + other for k, v in self.items()})

        result = self.copy()
        for key, value in items:
            result[key] = (result[key] + value) if key in self else value
        return QuantityDict(result)

    def __radd__(self, other: _OType) -> Self:
        return self + other

    def __mul__(self, other: _OType) -> Self:
        try:
            other.items()
        except AttributeError:
            return QuantityDict({k: v * other for k, v in self.items()})

        result = {k: v * other[k] for k, v in self.items() if k in other}
        return QuantityDict(result)

    def __rmul__(self, other: _OType) -> Self:
        return self * other

    def __pos__(self) -> Self:
        return self

    def __neg__(self) -> Self:
        return QuantityDict({k: -v for k, v in self.items()})

    def __sub__(self, other: _OType) -> Self:
        return self + (-other)

    def __rsub__(self, other: _OType) -> Self:
        return (-self) + other

    def __truediv__(self, other: _OType) -> Self:
        try:
            other.items()
        except AttributeError:
            return QuantityDict({k: v / other for k, v in self.items()})

        try:
            result = {k: v / other[k] for k, v in self.items()}
        except KeyError:
            msg = "Missing denominator element in QuantityDict division"
            raise ZeroDivisionError(msg) from None
        return QuantityDict(result)

    def __rtruediv__(self, other: _OType) -> Self:
        try:
            items = other.items()
        except AttributeError:
            return QuantityDict({k: other / v for k, v in self.items()})

        try:
            result = {k: v / self[k] for k, v in items}
        except KeyError:
            msg = "Missing denominator element in QuantityDict division"
            raise ZeroDivisionError(msg) from None
        return QuantityDict(result)

    def __matmul__(self, other: Self) -> Quantity:
        """Can assume other to be QuantityDict, otherwise use
        *-operator instead."""
        return (self * other).sum()

    def __rmatmul__(self, other) -> Quantity:
        """For 2 vectors, scalar product is commutative"""
        return other @ self

    def __pow__(self, other: _OType) -> Self:
        try:
            items = other.items()
        except AttributeError:
            result = {k: qpow(v, other) for k, v in self.items()}
        else:
            result = {k: qpow(v, other.get(k, 0)) for k, v in self.items()}
            for k, v in items:
                if k not in self:
                    result[k] = qpow(0.0, v)
        return QuantityDict(result)

    def __rpow__(self, other: _OType) -> Self:
        try:
            items = other.items()
        except AttributeError:
            result = {k: qpow(other, v) for k, v in self.items()}
        else:
            result = {k: qpow(other.get(k, 0) , v) for k, v in self.items()}
            for k, v in items:
                if k not in self:
                    result[k] = qpow(v, 0.0)
        return QuantityDict(result)

    def sum(self) -> Quantity:
        """Sum all elements of the QuantityDict object. Naturally, all
        elements must have equal physical dimensions

        >>> a = QuantityDict({
        ...         "B": Quantity("1 m"),
        ...         "C": Quantity("50 cm")})
        >>> print(f"{a.sum():~}")
        1.5 m
        """
        return sum(self.values())  # type: ignore[type-error]


_V = TypeVar("_V", float, Quantity, Mapping[str, Quantity])
UNFUNC_TYPE = Callable[[_V], _V]


def unary_func(quantity: _V, func: UNFUNC_TYPE) -> _V:
    """Call a unary function that requires a dimensionless argument on the
    argument, accepting that argument to be a float, a scalar quantity,
    a symbolic quantity, or a QuantityDict object (well, any dictionary
    of the above, actually)."""
    def scalar_func(value):
        """Call the unary function on the value's magnitude"""
        try:
            value = value.to_base_units()
            magnitude = value.magnitude
        except AttributeError:
            magnitude = value
        else:
            if not value.dimensionless:
                raise DimensionalityError(value.units, "dimensionless")
        return Quantity(func(magnitude))

    try:
        items = quantity.items()
    except AttributeError:
        return scalar_func(quantity)
    return QuantityDict({k: unary_func(v, func) for k, v in items})


def sqrt(quantity: _V) -> _V:
    """The square root function is a special case of a unary function in that
    the argument is not required to be dimensionless.

    >>> a = QuantityDict({
    ...         "B": Quantity("1 m**2"),
    ...         "C": Quantity("2500 cm**2")})
    >>> print(sqrt(a))
    {'B': <Quantity(1.0, 'meter')>, 'C': <Quantity(50.0, 'centimeter')>}
    """
    try:
        items = quantity.items()
    except AttributeError:
        return qsqrt(quantity)
    return QuantityDict({k: qsqrt(v) for k, v in items})


def log(quantity: _V) -> _V:
    """Determine natural logarithms, considering units of
    measurements. The main intent is to use this version for symbolic
    quantities and QuantityDict objects, but it also works on floats.

    >>> x = Quantity(10.0, "cm/m")
    >>> log(x)
    <Quantity(-2.30258509, 'dimensionless')>

    >>> a = {"A": SymbolQuantity("A", "dimless"),
    ...      "B": SymbolQuantity("B", "dimless")}
    >>> y = log(a)
    >>> for key, value in y.items(): print(f"{key}: {value:~}")
    A: log(A)
    B: log(B)

    >>> log(10)
    <Quantity(2.30258509, 'dimensionless')>

    The other unary functions are defined in the same manner.
    """
    return unary_func(quantity, cas.log)


log10: UNFUNC_TYPE = lambda x: unary_func(x, cas.log10)
exp: UNFUNC_TYPE = lambda x: unary_func(x, cas.exp)
sin: UNFUNC_TYPE = lambda x: unary_func(x, cas.sin)
cos: UNFUNC_TYPE = lambda x: unary_func(x, cas.cos)
tan: UNFUNC_TYPE = lambda x: unary_func(x, cas.tan)
arcsin: UNFUNC_TYPE = lambda x: unary_func(x, cas.arcsin)
arccos: UNFUNC_TYPE = lambda x: unary_func(x, cas.arccos)
arctan: UNFUNC_TYPE = lambda x: unary_func(x, cas.arctan)
sinh: UNFUNC_TYPE = lambda x: unary_func(x, cas.sinh)
cosh: UNFUNC_TYPE = lambda x: unary_func(x, cas.cosh)
tanh: UNFUNC_TYPE = lambda x: unary_func(x, cas.tanh)
arcsinh: UNFUNC_TYPE = lambda x: unary_func(x, cas.arcsinh)
arccosh: UNFUNC_TYPE = lambda x: unary_func(x, cas.arccosh)
arctanh: UNFUNC_TYPE = lambda x: unary_func(x, cas.arctanh)


def parse_quantities_in_struct(struct: Union[NestedMap[str], str])\
        -> Union[Quantity, NestedMap[Quantity]]:
    """Return a new struct that contains parsed quantities at the leaf
    values of the given input structure.

    The structure can be a nested dictionary, given the keys as strings and the
    leaf values as strings that can be parsed as ``pint`` quantities.
    For example:

    >>> from pprint import pprint
    >>> y = parse_quantities_in_struct({
    ...    'speed': {
    ...        'car': '100 km/hr',
    ...        'snail': '1 cm/min',
    ...        'fingernail': '1.2 mm/day'},
    ...    'weight': {
    ...        'car': '1.5 t',
    ...        'snail': '10 g',
    ...        'fingernail': '300 mg'}
    ... })
    >>> pprint(y)
    {'speed': {'car': <Quantity(100.0, 'kilometer / hour')>,
               'fingernail': <Quantity(1.2, 'millimeter / day')>,
               'snail': <Quantity(1.0, 'centimeter / minute')>},
     'weight': {'car': <Quantity(1.5, 'metric_ton')>,
                'fingernail': <Quantity(300, 'milligram')>,
                'snail': <Quantity(10, 'gram')>}}
    """
    try:
        items = struct.items()
    except AttributeError:
        return Quantity(struct)
    return {key: parse_quantities_in_struct(value) for key, value in items}


def quantity_dict_to_strings(struct: Quantity | NestedMap[Quantity],
                             significant_digits: int = 17) \
        -> str | NestedMap[str]:
    """Return a new structure with the quantity instances replaced by a string
    representation that is parsable by the :class:`simu.Quantity` constructor.

    Example:

    >>> from pprint import pprint
    >>> from simu import Quantity
    >>> struct = {'speed': {'car': Quantity(400 / 3, 'kilometer / hour'),
    ...                     'fingernail': Quantity(1.2, 'millimeter / day'),
    ...                     'snail': Quantity(1.0, 'centimeter / minute')},
    ...           'weight': {'car': Quantity(1.5, 'metric_ton'),
    ...                      'fingernail': Quantity(300, 'milligram'),
    ...                      'snail': Quantity(10, 'gram')}}
    >>> pprint(quantity_dict_to_strings(struct))
    {'speed': {'car': '133.33333333333334 km / h',
               'fingernail': '1.2 mm / d',
               'snail': '1 cm / min'},
     'weight': {'car': '1.5 t', 'fingernail': '300 mg', 'snail': '10 g'}}

    """
    try:
        items = struct.items()
    except AttributeError:
        return f"{struct:.{significant_digits}g~}"
    return {key: quantity_dict_to_strings(value, significant_digits)
            for key, value in items}




def extract_sub_structure(source: NestedMap[Quantity],
                          structure: NestedMap[str]) -> NestedMap[Quantity]:
    """Given a nested structure map ``structure`` that defines the units of
    measurement of leaf value quantities, extract those quantities from a source
    structure ``source``. A ``KeyError`` is raised if the source structure does
    not contain the requested data, and a ``DimensionalityError`` is raised if
    the queried quantity is of incompatible dimensions.

    Example:

    >>> from simu import Quantity
    >>> src = {"a": {"b": Quantity(1, "km")},
    ...        "c": Quantity(3, "degC"),
    ...        "d": {"e": Quantity(2, "s"), "f": Quantity(3, "kJ")}}
    >>> struct = {"a": {"b": "m"}, "d": {"e": "s"}}
    >>> print(extract_sub_structure(src, struct))
    {'a': {'b': <Quantity(1, 'kilometer')>}, 'd': {'e': <Quantity(2, 'second')>}}
    """
    def prepare(name: str, key: str, query: NestedMap[str],
                src: NestedMap[Quantity]) -> NestedMap[Quantity]:
        name = f"{name}.{key}" if name else key
        try:
            items = query.items()
        except AttributeError:
            try:
                src[key].to(query)
            except DimensionalityError as err:
                err.extra_msg = f" - Error fetching thermo parameter '{name}'."
                raise err from None
            return src[key]
        return {k: prepare(name, k, q, src[key]) for k, q in items}
    return {k: prepare("", k, s, source) for k, s in structure.items()}




from dataclasses import dataclass
from collections.abc import Iterator
from casadi import DM

from simu.core.utilities.quantity import Quantity
from simu.core.utilities.types import Map
from simu.core.utilities.errors import DimensionalityError


@dataclass
class Residual:
    """Class representing a single residual.

    The post-constructor will raise a ``DimensionalityError`` if the
    ``tolerance`` unit is not compatible with the ``value`` unit.

    A possible offset in the tolerance unit, such as in degrees Celsius or gauge
    pressures, will be eliminated automatically. That is:

    >>> from simu import SymbolQuantity as S, Quantity as Q
    >>> val = S("r", "barg")
    >>> r1 = Residual(val, Q(1e-7, "barg"))
    >>> r2 = Residual(val, Q(1e-7, "bar"))
    >>> f"{r1.tolerance - r2.tolerance:.15f~}"
    '0.000000000000000 bar'

    *(there is a silly rounding error, hence the trick with the formatting to
    make this doctest work)*
    """
    value: Quantity
    """The value as a symbolic quantity"""
    tolerance: Quantity
    """The absolute tolerance of the residual as a numeric quantity in the same
    physical dimensions as the value"""

    def __post_init__(self):
        tol_unit = self.tolerance.units
        try:
            self.value.to(tol_unit)
        except DimensionalityError:
            msg = f"Incompatible tolerance unit in residual"
            raise DimensionalityError(self.value.units, tol_unit, extra_msg=msg)

        # eliminate impact of offset in units like degC and barg
        self.tolerance -= Quantity(0.0, tol_unit)


ResidualProxy = Map[Residual]
"""This is just a dictionary, mapping the names (*str*) to the :class:`Residual`
objects"""

class ResidualHandler(ResidualProxy):
    """This class, being instantiated as the :attr:`simu.Model.residuals`
    attribute, allows to define residuals, i.e. process constraints."""

    def __init__(self):
        self.__residuals = {}

    def add(self, name: str, residual: Quantity,
            tol_unit: str, tol: float = 1e-7):
        """Define a residual, to approach zero withing its tolerance when the
        model is solved.

        :param name: A unique name within the local context
        :param residual: The residual to be zero.
        :param tol_unit: The unit of the tolerance, which must be compatible
          with the unit of the residual quantity.
        :param tol: The tolerance value. A normal recommendation is to give
          ``1e-7`` for a typical unit of measurement, which for extensive
          variables is largely problem size dependent. An industrial flow is
          typically characterized by ``kg/h``, while a lab scale experiment
          can operate with ``g/h``.
        """
        if name in self.__residuals:
            raise KeyError(f"Residual {name} already defined")
        self.__residuals[name] = Residual(residual, Quantity(DM(tol), tol_unit))

    def __getitem__(self, key: str) -> Residual:
        return self.__residuals[key]

    def __len__(self) -> int:
        return len(self.__residuals)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__residuals)

    def create_proxy(self) -> ResidualProxy:
        """Create a proxy object"""
        return self



"""Module defining classes related to thermodynamic state representation"""
# stdlib modules
from typing import Self
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass

from pint import DimensionalityError

# internal modules
from ..utilities import Quantity, qvertcat
from ..utilities.types import Map, NestedMap
from .species import SpeciesDefinition

@dataclass
class InitialState:
    """Dataclass describing an initial state, which is always defined in terms
    of temperature, pressure, and molar quantities.

    Temperature and pressure are scalar quantities of respective physical
    dimensions. the ``mol_vector`` quantity is vectorial and can arbitrarily
    be defined in units compatible with ``mol`` or ``mol/s``.
    The stored value will however be as a state, not a flow, i.e. compatible to
    ``mol``.
    """

    temperature: Quantity
    pressure: Quantity
    mol_vector: Quantity

    def __post_init__(self):
        """Check that attributes have compatible units, and convert mole flows
        to mole quantities (per 1 second).
        Raises ``DimensionalityError`` if dimensions do not match.
        """
        self.temperature.to("K")  # compatible unit?
        self.pressure.to("Pa")  # compatible unit?
        try:
            self.mol_vector.to("mol")
        except DimensionalityError:
            n = self.mol_vector.to("mol/s")
            self.mol_vector =  Quantity(n.magnitude, "mol")
        self.__setattr__ = self.__setattr_prototype  # freeze this object

    def __setattr_prototype(self, name, value):
        raise AttributeError(f"Cannot modify frozen instance: {name}")

    @classmethod
    def from_si(cls, temperature: float, pressure: float,
                mol_vector: Sequence[float]) -> Self:
        """Construct an initial state based on SI units, i.e. K, Pa and mol."""
        return cls(temperature=Quantity(temperature, "K"),
                   pressure=Quantity(pressure, "Pa"),
                   mol_vector=Quantity(mol_vector, "mol"))

    @classmethod
    def from_cbar(cls, temperature: float, pressure: float,
                  mol_vector: Sequence[float]) -> Self:
        """Construct an initial state based on degC, bar and mol as units."""
        return cls(temperature=Quantity(temperature, "degC"),
                   pressure=Quantity(pressure, "bar"),
                   mol_vector=Quantity(mol_vector, "mol"))

    @classmethod
    def from_std(cls, num_species: int):
        """Construct an initial state at 25 degC, 1 bar and one mol for each
        species."""
        return cls(temperature=Quantity(25, "degC"),
                   pressure=Quantity(1, "atm"),
                   mol_vector=Quantity([1.0] * num_species, "mol"))

    @classmethod
    def from_dict(cls, struct: NestedMap[Quantity], species: Sequence[str]):
        """Convert a nested Quantity map as obtained by :meth:`to_dict` to an
        initial state. The map must contain the top-level keys ``T``, ``p`` and
        ``n``, and a mapping from species to molar quantities as value of the
        ``n`` key. The correct sequence of ``species`` is provided separately,
        to be consistent with the targeted thermodynamic model.

        The method raises a ``KeyError``, if the elements of ``struct`` are not
        as expected, in particular considering the given ``species`` argument.
         """
        for s in struct["n"].keys():
            if s not in species:
                raise KeyError(f"Species {s} not expected")
        mole_list = [struct["n"][s] for s in species]
        units = mole_list[0].units
        mole_vector = [m.magnitude for m in mole_list]
        return cls(temperature=struct["T"],
                   pressure=struct["p"],
                   mol_vector=Quantity(mole_vector, units))


    def to_dict(self, species: Sequence[str]) -> NestedMap[Quantity]:
        """Convert initial state into dictionary of quantities.
        The mole vector is described as a sub-dictionary, mapping species
        to partial molar quantities"""
        mag = self.mol_vector.magnitude
        unit = self.mol_vector.units
        if len(species) != mag.size:
            raise ValueError("Incompatible length of species list")
        moles = {s: Quantity(mag[i], unit) for i, s in enumerate(species)}
        return {"T": self.temperature, "p": self.pressure, "n": moles}


class StateDefinition(ABC):
    """This class defines the interpretation of the state vector in terms of
    physical properties. This interpretation is then consumed by the
    contributions as input for their calculations towards the complete
    thermodynamic model."""

    @abstractmethod
    def prepare(self, result: dict, flow: bool = False):
        """This method can assume to find the state vector ``x`` in the
        ``result`` dictionary, and is expected to add the physical
        interpretation of its elements to the same dictionary. It is entirely
        up to the contributions that rely on this state.

        For the Gibbs state, the new elements would be ``T``, ``p``, and ``n``,
        denoting temperature, pressure and quantities respectively.

        The parameter ``flow`` impacts the definition of units of measurement.
        When ``True``, all extensive variables are divided by time.
        """
        ...

    def declare_vector_keys(
            self, species: Map[SpeciesDefinition]) -> Map[Sequence[str]]:
        """Declare the keys of vectorial state properties. In most cases,
        this will be the species names for the mole vector."""
        return {}

    @abstractmethod
    def reverse(self, state: InitialState) -> Sequence[float]:
        """Return the state vector as complete as possible with given
        temperature, pressure and quantities. The task of the contributions'
        :meth:`ThermoContribution.initial_state` method is it then to
        complete it. Missing elements shall be filled with None.
        """
        ...



"""This module defines types of complex data structures"""

from typing import Union, TypeVar, Self
from collections.abc import Mapping, MutableMapping

__V = TypeVar("__V")
"""An arbitrary type"""

Map = Mapping[str, __V]
"""A mapping of strings to another type"""

MutMap = MutableMapping[str, __V]
"""A mutable mapping of strings to another type"""

NestedMap = Map[__V | Self]
"""A nested mapping of strings to another type"""

NestedMutMap = MutMap[__V | Self]
"""A nested mutable mapping of strings to another type"""



from pytest import raises

from simu.core.utilities import Quantity
from simu.core.utilities.errors import DimensionalityError
from simu.core.utilities.residual import ResidualHandler


def test_create_residuals():
    handler = ResidualHandler()
    res = Quantity("3 m")   # this could never work, but not important here
    handler.add("Res 1", res, "km")
    handler.add("Res 2", res, "cm", tol=1e-5)
    tol = handler["Res 2"].tolerance
    assert f"{tol:~}" == "1e-05 cm"


def test_create_residual_inconsistent_unit():
    handler = ResidualHandler()
    res = Quantity("3 m")  # this could never work, but not important here
    with raises(DimensionalityError):
        handler.add("My Residual", res, "kJ")


def test_create_residual_same_name():
    handler = ResidualHandler()
    res = Quantity("3 m")  # this could never work, but not important here
    handler.add("My Residual", res, "km")
    with raises(KeyError):
        handler.add("My Residual",  2 * res, "cm")


from .thermo.contributions import (
    H0S0ReferenceState, LinearHeatCapacity, StandardState, IdealMix,
    GibbsIdealGas, HelmholtzIdealGas, ConstantGibbsVolume, MolecularWeight,
    NonSymmetricMixingRule, LinearMixingRule, BostonMathiasAlphaFunction,
    CriticalParameters, RedlichKwongEOSLiquid, RedlichKwongEOSGas,
    RedlichKwongAFunction, RedlichKwongBFunction, RedlichKwongMFactor,
    VolumeShift, Derivative, ChargeBalance, all_contributions)

from .thermo.state import all_states, HelmholtzState, GibbsState



# -*- coding: utf-8 -*-
"""Test module for cubic EOS contributions"""

# external modules
from numpy import linspace
from pytest import mark, raises
from matplotlib import pyplot

# internal modules
from simu.core.utilities import (
    assert_reproduction, base_unit, ParameterDictionary, SymbolQuantity,
    jacobian, QFunction, Quantity as Q, base_magnitude, qsum, unit_registry)
from simu.core.utilities.constants import R_GAS
from simu.core.thermo import InitialState
from simu.app.thermo.contributions import (
    CriticalParameters, LinearMixingRule, RedlichKwongEOSLiquid,
    RedlichKwongEOSGas, NonSymmetricMixingRule, RedlichKwongAFunction,
    RedlichKwongBFunction, RedlichKwongMFactor, VolumeShift)
from simu.app.thermo.contributions.cubic.rk import RedlichKwongEOS


# auxiliary functions
def sym(name: str, units: str) -> SymbolQuantity:
    """Define a scalar symbolic quantity"""
    return SymbolQuantity(name, base_unit(units))


def vec(name: str, size: int, units: str) -> SymbolQuantity:
    "define a vector symbolic quantity"
    return SymbolQuantity(name, base_unit(units), size)


def test_critical_parameters(species_definitions_ab):
    """Test definition of CriticalParameters contribution"""
    res = {}
    par = ParameterDictionary()
    cont = CriticalParameters(species_definitions_ab, {})
    cont.define(res)
    assert_reproduction(res)


def test_volume_shift(species_definitions_ab):
    """Test definition of VolumeShift contribution"""
    res = {}
    par = ParameterDictionary()
    cont = VolumeShift(species_definitions_ab, {})
    cont.define(res)
    assert_reproduction([res, cont.parameters])


def test_linear_mixing_rule(species_definitions_ab):
    """Test definition of LinearMixingRule contribution"""
    res = {"T": sym("T", "K"), "n": vec("n", 2, "mol"),
           "c_i": vec("c_i", 2, "m**3/mol")}
    par = ParameterDictionary()
    opt = {"target": "c"}
    cont = LinearMixingRule(species_definitions_ab, opt)
    cont.define(res)
    assert_reproduction(res["c"])


def test_redlich_kwong_eos(species_definitions_ab):
    """Test definition of RedlichKwongEOS contribution"""
    res = {
        "T": sym("T", "K"),
        "V": sym("V", "m**3"),
        "n": vec("n", 2, "mol"),
        "S": sym('S', "J/K"),
        "p": sym('p', "Pa"),
        "mu": vec('mu', 2, "J/mol")
    }
    res.update({
        "_ceos_a":
        sym("A0", "Pa*m**6") + res["T"] * sym('dAdT', "Pa*m**6/K"),
        "_ceos_b":
        sym("B0", "m**3") + res["T"] * sym('dBdT', "m**3/K"),
        "_ceos_c":
        sym("C0", "m**3"),
        "_state":
        vec("x", 4, "dimless")
    })
    cont = RedlichKwongEOSLiquid(species_definitions_ab, {})
    cont.define(res)
    keys = "S p mu _ceos_a_T _ceos_b_T".split()
    result = {k: res[k] for k in keys}
    assert_reproduction(result)


def test_abstract_class_init(species_definitions_ab):
    """Check that abstract RedlichKwongEOS class cannot be instantiated"""
    with raises(TypeError) as exception_info:
        # pylint: disable=abstract-class-instantiated
        RedlichKwongEOS(species_definitions_ab, {})
    assert "abstract" in str(exception_info.value)


def test_non_symmetric_mixing_rule(species_definitions_abc):
    """Test definition of NonSymmetricMixingRule contribution"""
    res = {
        "T": sym("T", "K"),
        "n": vec("n", 3, "mol"),
        "a_i": vec("a_i", 3, "Pa*m**6/mol")
    }
    options = {
        "k_1": [["A", "B"], ["A", "C"]],
        "k_2": [["A", "B"]],
        "l_1": [["B", "A"], ["C", "B"]],
        "target": "a"
    }
    cont = NonSymmetricMixingRule(species_definitions_abc, options)
    par = ParameterDictionary()
    cont.define(res)
    assert_reproduction(res["a"])


def test_non_symmetric_mixing_rule_no_interaction(species_definitions_abc):
    """Test definition of NonSymmetricMixingRule contribution"""
    res = {
        "T": sym("T", "K"),
        "n": vec("n", 3, "mol"),
        "a_i": vec("a_i", 3, "Pa*m**6/mol")
    }
    options = {
        "k_1": [],
        "k_2": [],
        "l_1": [],
        "target": "a"
    }
    cont = NonSymmetricMixingRule(species_definitions_abc, options)
    par = ParameterDictionary()
    cont.define(res)


def test_redlich_kwong_a_function(species_definitions_ab):
    """Test definition of RedlichKwongAFunction contribution"""
    res = {
        "_alpha": vec('alpha', 2, "dimless"),
        "_T_c": vec('T_c', 2, "K"),
        "_p_c": vec('p_c', 2, "bar")
    }
    cont = RedlichKwongAFunction(species_definitions_ab, {})
    cont.define(res)
    assert_reproduction(res["_ceos_a_i"])


def test_redlich_kwong_b_function(species_definitions_ab):
    """Test definition of RedlichKwongBFunction contribution"""
    res = {"_T_c": vec('T_c', 2, "K"), "_p_c": vec('p_c', 2, "bar")}
    cont = RedlichKwongBFunction(species_definitions_ab, {})
    cont.define(res)
    assert_reproduction(res["_ceos_b_i"])


def test_rk_m_factor(species_definitions_ab):
    """Test definition of RedlichKwongMFactor contribution"""
    res = {"_omega": vec('w', 2, "dimless")}
    cont = RedlichKwongMFactor(species_definitions_ab, {})
    cont.define(res)
    assert_reproduction(res["_m_factor"])


def test_boston_mathias_alpha_function(boston_mathias_alpha_function):
    """Test definition of BostonMathiasAlphaFunction contribution"""
    res, par = boston_mathias_alpha_function
    assert_reproduction(res["_alpha"][0])


def test_boston_mathias_alpha_function_smoothness(boston_mathias_alpha_function):
    """Check smoothness of alpha function at critical temperature, where
    the expression switches to the super-critical extrapolation"""
    res, par = boston_mathias_alpha_function

    args = {k: res[k] for k in ["T", "_T_c", "_m_factor"]}
    args["_eta"] = par.get_vector_quantity("eta")

    result = {"_alpha": res["_alpha"]}
    result["_a_t"] = jacobian(result["_alpha"], args["T"])
    result["_a_tt"] = jacobian(result["_a_t"], args["T"])

    f = QFunction(args, result)

    # now the numbers
    args = {
        "_T_c": Q([300, 400], "K"),
        "_m_factor": Q([0.6, 0.6]),
        "_eta": Q([0.12, 0.06])
    }

    def props(eps):
        args["T"] = args["_T_c"][0] + eps
        res = f(args)
        return {k: v[0] for k, v in res.items()}

    eps = Q(1e-10, "K")
    sub = props(-1.0 * eps)  # sub-critical
    sup = props(eps)  # super-critical

    assert abs(sub["_alpha"] - 1) < 1e-7, "sub-critical alpha unequal unity"
    assert abs(sup["_alpha"] - 1) < 1e-7, "super-critical alpha unequal unity"
    res = abs(sup["_a_t"] / sub["_a_t"] - 1)
    assert res < 1e-7, "first derivative not smooth"
    res = abs(sup["_a_tt"] / sub["_a_tt"] - 1)
    assert res < 1e-7, "second derivative not smooth"


def test_initialise_rk(species_definitions_ab):
    """Test volume initialisation of RK-model"""
    T, p = Q("100 degC"), Q("1 bar")
    n = Q([0.5, 0.5], "mol")
    # try to imitate water
    res = {
        "_ceos_a": Q("15 Pa * m**6"),
        "_ceos_b": Q("25 ml"),
        "_ceos_c": Q("10 ml")
    }
    liq = RedlichKwongEOSLiquid(species_definitions_ab, {})
    gas = RedlichKwongEOSGas(species_definitions_ab, {})
    ini_state = InitialState(temperature=T, pressure=p, mol_vector=n)
    v_liq = liq.initial_state(ini_state, res)[1]
    v_gas = gas.initial_state(ini_state, res)[1]
    assert abs(v_liq - 1.526e-5) < 1e-8  # value 1.5... is validated
    assert 0.02 < v_gas < 0.03


@mark.parametrize("cls", [RedlichKwongEOSGas, RedlichKwongEOSLiquid])
def test_initialise_rk2(cls, species_definitions_ab):
    """test initialisation of rk gas and liquid"""

    # define upstream expected results
    res = {
        "T": sym("T", "K"),
        "V": sym("V", "m**3"),
        "n": vec("n", 2, "mol"),
        "_ceos_a": sym("A", "bar*m**6"),
        "_ceos_b": sym("B", "m**3"),
        "_ceos_c": sym("C", "m**3"),
        "_state": vec('x', 4, "dimless")
    }
    ideal = {
        "S": sym("S", "J/K"),
        "mu": vec("mu", 2, "J/mol"),
        "p": sym("p", "Pa")
    }

    res.update(ideal)
    cont = cls(species_definitions_ab, {})
    cont.define(res)  # now the ideal part in res is overwritten

    # now with number quantities
    T, p, n = Q("100 degC"), Q("1 bar"), Q([0.5, 0.5], "mol")
    # try to imitate water
    res_num = {
        "T": T,
        "n": n,
        "_ceos_a": Q("15 Pa * m**6"),
        "_ceos_b": Q("25 ml"),
        "_ceos_c": Q("10 ml"),
    }
    ideal_num = {"S": Q("0 J/K"), "p": Q("0 Pa"), "mu": Q([0, 0], "J/mol")}
    ini_state = InitialState(temperature=T, pressure=p, mol_vector=n)
    state = cont.initial_state(ini_state, res_num)

    # is the rest of the state (except volume) reproduced?
    assert base_magnitude(T) == state[0]
    assert base_magnitude(n).tolist() == state[2:]
    res_num["V"] = Q(state[1], "m**3")

    # calculate contribution values with initial state to reproduce pressure
    args = {k: res[k] for k in res_num}
    args.update(ideal)
    func = QFunction(args, {"p_res": res["p"]})

    args = {**res_num, **ideal_num}
    p2 = func(args)["p_res"] + sum(n) * R_GAS * T / args["V"]
    assert abs(p2 - p) < Q("1 Pa")


# *** auxiliary routines (TODO: plot should be an example instead!)

def plot_pv(res):
    """auxiliary method to plot pv-graph and linear/quadratic approximation"""
    T, V = res["T"], res["V"]

    def p(V):
        A, B, C = [res[i] for i in "_ceos_a _ceos_b _ceos_c".split()]
        A /= 33.7
        VC = V + C
        return qsum(res["n"]) * R_GAS * T / (VC - B) - A / VC / (VC + B)

    # only plot if running this file interactively
    volumes = linspace(45, 52, num=100) * Q("1 ml")
    pressures = p(volumes)
    lin_p = p(V) + (volumes - V) * res["_dp_dV"]
    ddp_dv2 = res["_ddp_dV_dx"][1] * Q("m**-3")
    quad_p = lin_p + (volumes - V)**2 * ddp_dv2 / 2

    unit_registry.setup_matplotlib(True)  # allow plotting with units
    pyplot.plot(volumes, pressures.to("bar"), 1)
    pyplot.plot(volumes, lin_p.to("bar"), ":")
    pyplot.plot(volumes, quad_p.to("bar"), "--")
    pyplot.xlim([volumes[0], volumes[-1]])
    # pyplot.ylim(Q([0, 15], "bar"))
    pyplot.xlabel("V [m3]")
    pyplot.ylabel("p [bar]")
    pyplot.grid()
    pyplot.show()


if __name__ == "__main__":
    from pytest import main
    from sys import argv
    # only this file, very verbose and print stdout when started from here.
    main([__file__, "-v", "-v", "-s", "-rP"] + argv[1:])



from abc import ABC, abstractmethod
from typing import Optional
from collections.abc import Iterable, Collection, Mapping, Sequence

from ..utilities import (
    Quantity, SymbolQuantity, QuantityDict, QFunction, extract_sub_structure)
from ..utilities.types import Map, MutMap, NestedMap

from . import (
    ThermoFrame, InitialState, ThermoParameterStore, SpeciesDB,
    ThermoFactory, SpeciesDefinition)


class MaterialSpec:
    """Representation of a requirement to a material object.

    Objects of this class are used to define the requirements to a material,
    typically in a material port of a Model. It defines which species are
    required and whether additional species are allowed.
    """

    def __init__(self, species: Optional[Iterable[str]] = None,
                 flow: bool = True):
        """Create an instance based on the arguments as follows:

        :species: If ``None`` (default), allow any species. If containing
          a species name ``*``, allow any species, but demand the species
          that are element of the given iterable. If  ``species`` does not
          contain ``*``, lock the specification to the given set of species.
        :flow: ``True`` if the material needs to represent a flow, ``False`` if
          a state is to be represented.
        """
        self.__flow = flow
        self.__locked = not (species is None or "*" in species)
        self.__species = set() if species is None else set(species) - set("*")

    @classmethod
    @property
    def flow(cls):
        """A generic flow instance, allowing any species"""
        return cls()

    @classmethod
    @property
    def state(cls):
        """A generic state instance, allowing any species"""
        return cls(flow=False)

    @property
    def species(self) -> set[str]:
        """The set of species that must be provided"""
        return set(self.__species)

    @property
    def locked(self) -> bool:
        """Whether the specification allows other species than the ones
        listed as :attr:`species`."""
        return self.__locked

    def is_flow(self) -> bool:
        """Whether the specification is a flow (``True``) or
        a state (``False``)"""
        return self.__flow

    def is_compatible(self, material: "Material") -> bool:
        """Return true if the given material is compatible with this one.
        That is:

        - none of the specified species are missing in the material
        - if specification locks species set, none of the material
          species are missing in the specification
        - flows are only compatible with flows, states are only compatible
          with states.
        """
        spe, mspe = self.species, set(material.species)
        locked = self.locked
        flow_comp = self.is_flow() == material.is_flow()
        return flow_comp and not ((spe - mspe) or (locked and (mspe - spe)))


class Material(MutMap[Quantity]):
    """This class represents a material"""

    definition: "MaterialDefinition"
    """The underlying definition"""

    initial_state: InitialState
    """The provided initial state"""

    def __init__(self,
                 definition: "MaterialDefinition",
                 flow: bool):
        """Define a material based on the given definition. The ``flow`` flag
        indicates whether a *flow* or a *state* is to be generated."""
        self.definition = definition
        self.initial_state = definition.initial_state

        frame = definition.frame
        params = definition.store.get_symbols(frame.parameter_structure)
        self.__state = frame.create_symbol_state()
        props = frame(self.__state, params, squeeze_results=False, flow=flow)
        vectors = frame.vector_keys

        def convert(n: str, prop: SymbolQuantity) -> Quantity | QuantityDict:
            """If property is a registered vector, convert it to a QuantityDict,
            otherwise just return the quantity itself."""
            nonlocal vectors
            mag, unit = prop.magnitude, prop.units
            try:
                keys = vectors[n]
            except KeyError:
                return prop
            if mag.size() != (len(keys), 1):
                msg = f"Property {n} has improper shape: {mag.size()}, " \
                      f"should be ({len(keys)}, 1)"
                raise ValueError(msg)
            result = {s: Quantity(mag[i], unit) for i, s in enumerate(keys)}
            return QuantityDict(result)

        self.__properties = {n: convert(n, p) for n, p in props["props"].items()
                             if not n.startswith("_")}
        self.__bounds = props.get("bounds", {})
        self.__residuals = props.get("residuals", {})
        self.__normed_residuals = props.get("normed_residuals", {})
        self.__flow = flow

        # create a QFunction to map a state and parameters into a new initial
        # state
        args = {"state": Quantity(self.__state), "param": params}
        props = frame(self.__state, params, squeeze_results=False, flow=flow)
        res = {n: props["props"][n] for n in "Tpn"}
        self.__ini_func = QFunction(args, res, "ini_func")

    def retain_initial_state(self, state: Sequence[float],
                             parameters: NestedMap[Quantity]):
        param_struct = self.definition.frame.parameter_structure
        store_name = self.definition.store.name
        parameters = extract_sub_structure(parameters[store_name], param_struct)
        args = {"state": Quantity(state), "param": parameters}
        res = self.__ini_func(args)
        self.initial_state = InitialState(
            temperature=res["T"],
            pressure=res["p"],
            mol_vector=res["n"].reshape(-1))  # make sure it's a vector

    @property
    def species(self) -> Collection[str]:
        """The species names"""
        return self.definition.species

    @property
    def species_definitions(self) -> Map[SpeciesDefinition]:
        return self.definition.species_definitions

    @property
    def sym_state(self) -> Map[Quantity]:
        """A dictionary of quantities representing the state"""
        state = self.__state.nonzeros()
        return {f"x_{k:03d}": Quantity(x_k) for k, x_k in enumerate(state)}

    @property
    def bounds(self) -> NestedMap[Quantity]:
        """Return the bounds of the material as a nested mapping. The top
        level keys will be the names of the original contributions, while the
        second level keys will be the names of the bounds. The mapping will
        only have these two levels.
        """
        return self.__bounds

    def residuals(self, normed: bool = False) -> NestedMap[Quantity]:
        """Return the residuals of the material as a nested mapping. The top
        level keys will be the names of the original contributions, while the
        second level keys will be the names of the residuals. The mapping will
        only have these two levels.

        If ``normed`` is ``true``, return the dimensionless ratio of residual
        values and their tolerances instead.
        """
        return self.__normed_residuals if normed else self.__residuals


    def __getitem__(self, key: str) -> Quantity:
        """Return a (symbolic) property that is calculated by the underlying
        thermodynamic model or later supplements via :meth:`__setitem__`."""
        return self.__properties[key]

    def __setitem__(self, key: str, symbol: Quantity):
        """Supplement the material with another property. The ``symbol``
        argument must be a symbolic quantity, and it is highly immoral to supply
        a symbol that is a function of more than prior material properties.
        """
        if key in self.__properties:
            raise KeyError(f"Property '{key}' already exists in material")
        self.__properties[key] = symbol

    def __delitem__(self, key):
        raise TypeError("Property deletion is disabled to avoid opaque "
                        "property name interpretation.")

    def __iter__(self):
        return iter(self.__properties)

    def __len__(self):
        return len(self.__properties)

    def is_flow(self):
        """Return if the material represents a flow (``True``) or not
        (``False``)"""
        return self.__flow


class MaterialDefinition:
    """A ``MaterialDefinition`` object defines a material type by its

      - frame of thermodynamic contributions,
      - initial state
      - source of thermodynamic parameters
    """

    __frame: ThermoFrame
    __initial_state: InitialState
    __store: ThermoParameterStore

    def __init__(self, frame: ThermoFrame, initial_state: InitialState,
                 store: ThermoParameterStore):
        self.__frame = frame
        self.initial_state = initial_state
        store.get_symbols(frame.parameter_structure)  # trigger querying params
        self.__store = store

    @property
    def spec(self) -> MaterialSpec:
        """Return a material spec object that is implemented by this
        definition"""
        return MaterialSpec(self.frame.species)

    @property
    def frame(self) -> ThermoFrame:
        return self.__frame

    @property
    def store(self) -> ThermoParameterStore:
        return self.__store

    @property
    def species(self) -> Collection[str]:
        """The species names"""
        return self.frame.species

    @property
    def species_definitions(self) -> Map[SpeciesDefinition]:
        return self.frame.species_definitions

    @property
    def initial_state(self):
        return self.__initial_state

    @initial_state.setter
    def initial_state(self, new_state: InitialState):
        num_species = len(self.__frame.species)
        num_init_species = len(new_state.mol_vector.magnitude)
        if num_init_species != num_species:
            raise ValueError(
                f"Incompatible initial state with {num_init_species} "
                f"species, while {num_species} is/are expected"
            )
        self.__initial_state = new_state

    def create_flow(self) -> Material:
        return Material(self, True)

    def create_state(self) -> Material:
        return Material(self, False)


class MaterialLab:
    """A MaterialLab is an object to help design and define material
    definitions. It aggregates one or more ThermoParameterStore, the global
    SpeciesDefinition, and the ThermoFactory.

    A new material can be defined based on a species set, an initial state,
    and the contribution list.
    """
    def __init__(self, factory: ThermoFactory, species_db: SpeciesDB,
                 param_store: ThermoParameterStore):
        self.__factory = factory
        self.__species_db = species_db
        self.__param_store = param_store

    def define_material(self, species: Sequence[str],
                        initial_state: InitialState,
                        structure: Mapping) -> MaterialDefinition:
        """This method creates a material definition."""
        species_map = {s: self.__species_db[s] for s in species}
        frame = self.__factory.create_frame(species_map, structure)
        return MaterialDefinition(frame, initial_state, self.__param_store)



"""This module implements functionality concerning the numerical handling
of the top model instance."""

from typing import Optional
from collections.abc import Callable, Sequence, Collection
from copy import deepcopy
from casadi import vertcat, SX
from pint import Unit

from ..utilities import (
    flatten_dictionary, quantity_dict_to_strings,
    parse_quantities_in_struct, nested_map)
from ..utilities.types import NestedMap, NestedMutMap, Map, MutMap
from ..utilities.quantity import Quantity, QFunction, jacobian
from ..utilities.structures import FLATTEN_SEPARATOR
from ..utilities.errors import DataFlowError

from .base import ModelProxy
from ..thermo import ThermoParameterStore
from ... import unflatten_dictionary, InitialState


# TODO:
#  - set parameters and get parameters


class NumericHandler:
    """This class implements the function object describing the top level
    model."""
    THERMO_PARAMS: str = "thermo_params"
    MODEL_PARAMS: str = "model_params"
    MODEL_PROPS: str = "model_props"
    THERMO_PROPS: str = "thermo_props"
    RESIDUALS: str = "residuals"
    STATE_VEC: str = "states"
    RES_VEC: str = "residuals"
    BOUND_VEC: str = "bounds"
    VECTORS: str = "vectors"
    JACOBIANS: str = "jacobians"

    def __init__(self, model: ModelProxy, port_properties: bool = True):
        """The option ``port_properties`` determines whether the properties
        of connected materials are also reported from a child model's
        perspective by the name of their ports."""
        self.options = {
            "port_properties": port_properties
        }
        self.model = model
        # the name vectors of vector arguments
        self.__vec_arg_names: MutMap[Sequence[str]] = {}
        self.__vec_res_names: MutMap[Sequence[str]] = {}

        # the symbolic argument structure
        self.__symargs: NestedMutMap[Quantity] = self.__collect_arguments()
        # the symbolic result structure
        self.__symres: NestedMutMap[Quantity] = self.__collect_results()
        # the numerical argument structure with initial values
        self.__arguments: MutMap[Quantity] = {}

    @property
    def function(self) -> QFunction:
        """Create a Function object based on currently available argument
        and result structures."""
        return QFunction(self.__symargs, self.__symres, "model")

    def vector_arg_names(self, key: str) -> Sequence[str]:
        """Return the names for the argument vector of given ``key``"""
        return self.__vec_arg_names[key]

    def vector_res_names(self, key: str) -> Sequence[str]:
        """Return the names for the result vector of given ``key``"""
        return self.__vec_res_names[key]

    @property
    def arguments(self) -> Map[Quantity]:
        """The function arguments as numerical values. A DataFlowError is
        thrown, if not all numerical values are known.
        A deepcopy of the structure is provided, so the returned data can be
        altered without side effects.
        """
        if not self.__arguments:
            self.__arguments = self.__collect_argument_values()
        return deepcopy(self.__arguments)

    def export_state(self) -> NestedMutMap[str]:
        """Export the internal state of the model in a hierarchical structure,
        whereas all thermodynamic states are given in :math:`T, p, n`.

        As by the philosophy of the chosen approach, only the thermodynamic
        models know how to obtain their internal state from any :math:`T, p, n`
        specification.

        The returned structure is meant to be easy to store for instance in
        yaml or json format, and easy to edit. One can use
        :func:`~simu.parse_quantities_in_struct` to convert the values of the
        data structure into :class:`~simu.Quantity` objects for programmatic
        processing.

        """
        def fetch_initial_states(model: ModelProxy) -> MutMap[Quantity]:
            """fetch material states from a specific model"""
            mat_proxy = model.materials
            return {k: m.initial_state.to_dict(m.species)
                    for k, m in mat_proxy.handler.items()
                    if k not in mat_proxy}

        thermo =  self.__fetch(self.model, fetch_initial_states, "state")
        # TODO: when non-canonical states are implemented, collect them here.

        return quantity_dict_to_strings(
            {"thermo": thermo,
             "non-canonical": {}}
        )

    def import_state(self, state: NestedMap[str],
                     allow_missing: bool=False, allow_extra: bool = False)\
            -> NestedMap[str]:
        """Imports the state data in terms of :math:`T, p, n` as exported by
        :meth:`export_state`.

        :param state: A nested mapping as returned by :meth:`export_state`,
          containing the two first-level keys ``thermo`` (for thermodynamic
          states) and ``non-canonical`` for non-canonical states.
        :param allow_missing: If true, throw a ``ValueError`` if there are model
          states in the model that are not defined in the given ``state``
          structure.
        :param allow_extra: If false, throw a ``ValueError`` if there are states
          in the given ``state`` structure that are not present in the model.
        :return: A nested mapping of same structure as ``state``, but containing
          states that are not present in the model (value = ``extra``) and
          states that were not given as part of ``state`` (value = ``missing``)
        """
        def mk_new_path(path: str, name: str) -> str:
            name = name.replace(FLATTEN_SEPARATOR, rf"\{FLATTEN_SEPARATOR}")
            return name if not path else f"{path}{FLATTEN_SEPARATOR}{name}"

        def traverse(model: ModelProxy, state_part: NestedMap[Quantity],
                     path: str):
            # process local material objects
            all_names = set()
            for name, material in model.materials.handler.items():
                new_path = mk_new_path(path, name)
                all_names.add(name)
                try:
                    new_part = state_part[name]
                except KeyError:
                    if not allow_missing:
                        raise
                    result[new_path] = "missing"
                else:
                    material.initial_state = \
                        InitialState.from_dict(new_part, material.species)

            # traverse down into model hierarchy
            for name, proxy in model.hierarchy.handler.items():
                all_names.add(name)
                new_path = mk_new_path(path, name)
                try:
                    new_part = state_part[name]
                except KeyError:
                    if not allow_missing:
                        raise
                    result[new_path] = "missing"
                else:
                    traverse(proxy, new_part, new_path)

            # detect states that are not defined in model
            for name in state_part.keys():
                new_path = mk_new_path(path, name)
                if not name in all_names:
                    if not allow_extra:
                        msg = f"{new_path} not found in model structure"
                        raise KeyError(msg)
                    result[new_path] = "extra"

        self.__arguments = {}  # force reread
        result = {}
        traverse(self.model, parse_quantities_in_struct(state["thermo"]), "")
        return unflatten_dictionary(result)


    def retain_state(self, state: Sequence[float],
                     parameters: NestedMap[Quantity]):
        """Given a numeric ``state`` vector and the current set of
        ``parameters`` as a nested mapping of quantities, store the values for
        temperature, pressure and molar quantities back into the internal
        representations of the initial thermodynamic states.
        """
        def fetch_retain_initial_state(model: ModelProxy):
            """retain initial states for a specific model"""
            nonlocal index
            for m in model.materials.handler.values():
                state_length = 2 + len(m.species)
                state_part = state[index:index + state_length]
                m.retain_initial_state(state_part, parameters)
                index += state_length

        index = 0  # mutable index
        self.__arguments = {}  # force reread
        self.__traverse(self.model, fetch_retain_initial_state)

        # todo: if there are non-canonical states, treat them now.


    def extract_parameters(self, key: str,
                           definition: NestedMap[str]) -> Quantity:
        """collect the parameter symbols addressed by the ``definition``
        argument, which defines the unit of measurement for each parameter -
        as the numerical parameter vector elements must be considered
        dimensionless.

        .. note::

            Unfortunately, we cannot easily allow unit conversion (even
            compatible units) at this point, as the parameters are already
            independent variables used to build up the symbolic graph.
            Well, it can be done by first building the original function,
            and then call the function as f(c(x)), where c(x) is the unit
            conversion.

        These parameter symbols and values are then removed from the original
        argument structure and instead added to the vector entry as
        dimensionless entities.

        :param key: The name to be used for the parameter set
        :param definition: The parameters to be extracted
        """
        def traverse(parameters: NestedMap[str],
                     symbols: NestedMutMap[Quantity],
                     arguments: NestedMutMap[Quantity]) -> \
                (Sequence[str], Sequence[SX], Sequence[float]):
            """recursively go through struct, extract and remove both values
            and symbols."""
            try:
                items = parameters.items()
            except AttributeError:
                return None, None, None

            nams, syms, args = [], [], []
            for k, value in items:
                n, s, v = traverse(value, symbols[k], arguments[k])
                if s is None:
                    if symbols[k].units != Unit(value):
                        msg = "No unit conversion possible for parameter " \
                            f"{k}: from {symbols[k].units:~} to {value}."
                        raise ValueError(msg)
                    nams.append(k)
                    syms.append(symbols[k].magnitude)
                    args.append(arguments[k].magnitude)
                    del symbols[k]
                    del arguments[k]
                else:
                    nams.extend([f"{k}/{n_i}" for n_i in n])
                    syms.extend(s)
                    args.extend(v)
            return nams, syms, args

        if key in self.__symargs[NumericHandler.VECTORS]:
            msg = f"A parameter vector of name '{key}' is already used."
            raise KeyError(msg)

        if not self.__arguments:
            self.__arguments = self.__collect_argument_values()
        nam, sym, arg = traverse(definition, self.__symargs, self.__arguments)

        result = Quantity(vertcat(*sym))
        self.__symargs[NumericHandler.VECTORS][key] = result
        values = Quantity(arg)
        self.__arguments[NumericHandler.VECTORS][key] = values
        self.__vec_arg_names[key] = nam
        return result

    def collect_properties(self, key: str,
                           definition: NestedMap[str]) -> Quantity:
        """Collect the property symbols addressed by the ``definition``
        argument, which defines the unit of measurement for each property
        to be scaled with - as the numerical property vector elements must
        be dimensionless."""
        def traverse(properties: NestedMap[str],
                     symbols: NestedMap[Quantity]) -> \
                (Sequence[str], Sequence[SX]):
            """Recursively go through property struct, collect symbols, and
            convert them into desired units."""
            try:
                items = properties.items()
            except AttributeError:
                return None, None

            nams, syms = [], []
            for k, value in items:
                n, s = traverse(value, symbols[k])
                if n is None:
                    nams.append(k)
                    syms.append(symbols[k].to(value).magnitude)
                else:
                    nams.extend([f"{k}/{n_i}" for n_i in n])
                    syms.extend(s)
            return nams, syms

        if key in self.__symres[NumericHandler.VECTORS]:
            msg = f"A property vector of name '{key}' is already used."
            raise KeyError(msg)

        nam, sym = traverse(definition, self.__symres)

        result = Quantity(vertcat(*sym))
        self.__symres[NumericHandler.VECTORS][key] = result
        self.__vec_res_names[key] = nam
        return result

    def register_jacobian(self, dependent: str, independent: str) -> str:
        """Add the given symbols to the jacobian structure. These symbols must
        be a function of the arguments, or else the function cannot be
        created. The key must be unique.
        """
        dep = self.__symres[self.VECTORS][dependent]
        ind = self.__symargs[self.VECTORS][independent]
        jac = jacobian(dep, ind)
        key = f"d_({dependent})/d_({independent})"
        self.__symres[self.JACOBIANS][key] = jac
        return key

    def __collect_arguments(self) -> NestedMutMap[Quantity]:
        """Create a function that has the following arguments, each of them as
        a flat dictionary:

            - Material States
            - Model Parameters
            - Thermodynamic Parameters

        For child models, only the free parameters are collected.
        """
        mod = self.model
        fetch = self.__fetch
        to_vector = self.__to_vector

        def fetch_material_states(model: ModelProxy) -> MutMap[Quantity]:
            """fetch material states from a specific model"""
            mat_proxy = model.materials
            return {k: m.sym_state for k, m in mat_proxy.handler.items()
                    if k not in mat_proxy}

        def fetch_parameters(model: ModelProxy) -> MutMap[Quantity]:
            """fetch model parameters from a specific model"""
            return dict(model.parameters.free)

        def fetch_store_param(model: ModelProxy) -> NestedMap[Quantity]:
            """fetch thermodynamic parameters from the stores"""
            stores = self.__fetch_thermo_stores(model)
            names = {store.name for store in stores}
            if len(names) < len(stores):
                raise ValueError("When using multiple ThermoPropertyStores, "
                                 "they have to have unique names")
            return {store.name: store.get_all_symbols() for store in stores}

        states_struct = fetch(mod, fetch_material_states, "state")
        states, state_names = to_vector(states_struct)
        self.__vec_arg_names[self.STATE_VEC] = state_names

        return {
            self.THERMO_PARAMS: fetch_store_param(mod),
            self.MODEL_PARAMS: fetch(mod, fetch_parameters, "parameter"),
            self.VECTORS: {
                self.STATE_VEC: states,
            }
        }

    def __collect_results(self) -> NestedMutMap[Quantity]:
        """The result of the function consists of

            - Model Properties
            - Thermodynamic (state) properties
            - Residuals

        All the data is to be collected from the model and all child model
        proxies.
        """
        def fetch_residuals(model: ModelProxy,
                            normed: bool = False) -> NestedMutMap[Quantity]:
            """fetch residuals from a specific model"""
            def extract(entity):
                if normed:
                    return (entity.value / entity.tolerance).to("")
                return entity.value

            # find residuals of materials
            mat_proxy = model.materials
            res = {k: m.residuals(normed) for k, m in mat_proxy.handler.items()
                   if k not in mat_proxy}
            # add residuals of model (detect name clashes)
            clash = set(res.keys()) & set(model.residuals.keys())
            if clash:
                clash = ", ".join(clash)
                msg = f"Name clash of residuals and child modules: {clash}"
                raise ValueError(msg)

            res.update({k: extract(v) for k, v in model.residuals.items()})
            return res

        def fetch_bounds(model: ModelProxy) -> MutMap[Quantity]:
            mat_proxy = model.materials
            res = {k: m.bounds for k, m in mat_proxy.handler.items()
                   if k not in mat_proxy}
            clash = set(res.keys()) & set(model.bounds.keys())
            if clash:
                clash = ", ".join(clash)
                msg = f"Name clash of bounds and child modules: {clash}"
                raise ValueError(msg)
            res.update(model.bounds)
            return res

        def fetch_mod_props(model: ModelProxy) -> MutMap[Quantity]:
            """fetch model properties from a specific model"""
            return dict(model.properties.items())

        def fetch_thermo_props(model: ModelProxy) -> MutMap[Quantity]:
            """fetch properties of materials in a specific model"""
            ports = self.options["port_properties"]
            mat_proxy = model.materials
            return {k: v for k, v in mat_proxy.handler.items()
                    if ports or k not in mat_proxy}

        mod = self.model
        fetch = self.__fetch
        to_vector = self.__to_vector

        residual_structure = fetch(mod, lambda x: fetch_residuals(x, True),
                                   "normalised residual")
        bounds_structure = fetch(mod, fetch_bounds, "bound")
        residuals, residual_names = to_vector(residual_structure)
        bounds, bound_names = to_vector(bounds_structure)
        self.__vec_res_names[self.RES_VEC] = residual_names
        self.__vec_res_names[self.BOUND_VEC] = bound_names
        return {
            self.MODEL_PROPS: fetch(mod, fetch_mod_props, "model property"),
            self.THERMO_PROPS:
                fetch(mod, fetch_thermo_props, "thermo property"),
            self.RESIDUALS: fetch(mod, lambda x: fetch_residuals(x, False),
                                  "residual"),
            self.VECTORS: {
                self.RES_VEC: residuals,
                self.BOUND_VEC: bounds
            },
            self.JACOBIANS: {}
        }

        # The following jacobian is always needed
        # TODO: no it isn't!
        #   I might create 2 functions, one for a Newton step and one
        #   much cheaper to evaluate for line search.
        # maybe define flag in constructor whether to create this one right
        # away.

        # TODO: also might add one entry which is the mean square residual

        # if residuals.magnitude.rows() and states.magnitude.rows():
        #     self.__symres[cls.DR_DX] = jacobian(residuals, states)
        # else:
        #     self.__symres[cls.DR_DX] = Quantity(SX.sym("dr_dx", 0))
        #

    def __collect_argument_values(self) -> NestedMutMap[Quantity]:
        """Fetch initial states from materials, parameter values from
        thermo parameter stores, and parameter values from parameter handlers.
        """
        def fetch_states(model: ModelProxy) -> NestedMutMap[Quantity]:
            """Fetch the initial state variables from the materials of a
            specific model"""
            result = {}
            for k, m in model.materials.handler.items():
                init = m.initial_state
                try:
                    params = m.definition.store.get_all_values()
                except KeyError:
                    msg = "Missing values for thermodynamic parameters"
                    raise DataFlowError(msg)
                state = m.definition.frame.initial_state(init, params)
                dic = {f"x_{i:03d}": Quantity(x) for i, x in enumerate(state)}
                result[k] = dic
            return result

        def fetch_store_param() -> NestedMap[Quantity]:
            """fetch thermodynamic parameter values from the stores"""
            stores = NumericHandler.__fetch_thermo_stores(self.model)
            names = {store.name for store in stores}
            if len(names) < len(stores):
                raise ValueError("When using multiple ThermoPropertyStores, "
                                 "they have to have unique names")
            return {store.name: store.get_all_values() for store in stores}

        fetch = self.__fetch
        to_vector = self.__to_vector

        states = to_vector(fetch(self.model, fetch_states, "state"))[0]
        model_param = fetch(self.model, lambda m: m.parameters.values,
                            "parameter")
        return {
            self.VECTORS: {
                self.STATE_VEC: states,
            },
            self.MODEL_PARAMS: model_param,
            self.THERMO_PARAMS: fetch_store_param()
        }

    @staticmethod
    def __to_vector(struct: NestedMap[Quantity]) -> (Quantity, Sequence[str]):
        raw = [v.magnitude for v in flatten_dictionary(struct).values()]
        names = list(flatten_dictionary(struct).keys())
        return Quantity(vertcat(*raw)), names

    @staticmethod
    def __fetch(
            root: ModelProxy,
            func: Callable[[ModelProxy], NestedMutMap[Quantity]],
            typ: str,
            path: Optional[Sequence[str]] = None) -> NestedMutMap[Quantity]:
        """Drill recursively into child models to collect all data. The result
        is a nested dictionary, such that name clashes between child models and
        parameters are not permitted and will raise a ``ValueError``.
        """
        call_self = NumericHandler.__fetch
        if path is None:
            path = []
        result: NestedMutMap[Quantity] = func(root)
        for name, proxy in root.hierarchy.handler.items():
            if name in result:
                context = ".".join(path)
                msg = f"Child model / {typ} name clash:" \
                    f"'{name}' in {context}"
                raise ValueError(msg)
            result[name] = call_self(proxy, func, typ, path + [name])
        return result

    @staticmethod
    def __traverse(
            root: ModelProxy,
            func: Callable[[ModelProxy], None]):
        """Drill recursively into child modules to perform some action."""
        call_self = NumericHandler.__traverse
        func(root)
        for name, proxy in root.hierarchy.handler.items():
            call_self(proxy, func)

    @staticmethod
    def __fetch_thermo_stores(model: ModelProxy) \
            -> Collection[ThermoParameterStore]:
        call_self = NumericHandler.__fetch_thermo_stores
        result = {m.definition.store
                  for m in model.materials.handler.values()}
        for proxy in model.hierarchy.handler.values():
            result |= call_self(proxy)
        return result



"""The utilities module contains all required data structures and utility
functions that are not specific to particular objects, but have a certain
general purpose."""

from .structures import (
    MCounter, flatten_dictionary, unflatten_dictionary, nested_map)
from .quantity import (
    QFunction, Quantity, SymbolQuantity, base_magnitude, base_unit,
    conditional, jacobian, qpow, qvertcat, qsqrt, qsum, unit_registry,
    extract_units_dictionary, simplify_quantity)
from .qstructures import (
    ParameterDictionary, QuantityDict, parse_quantities_in_struct, exp, log,
    log10, sin, cos, tan, arcsin, arccos, arctan, sinh, cosh, tanh, arcsinh,
    arccosh, arctanh, sqrt, extract_sub_structure, quantity_dict_to_strings)
from .testing import assert_reproduction, user_agree



# stdlib
from typing import Iterable

# external
from yaml import safe_load

# internal
from . import all_contributions, all_states
from ..data import DATA_DIR
from ... import ThermoFactory


class ExampleThermoFactory(ThermoFactory):
    """This ThermoFactory subclass is capable of creating frames from the base
    SiMu installation, hence thermodynamic models that are found in open
    literature."""
    def __init__(self):
        """Default and only constructor"""
        ThermoFactory.__init__(self)
        self.register(*all_contributions)
        for state in all_states:
            self.register_state_definition(state)

        with open(DATA_DIR / "structures.yml", encoding='UTF-8') as file:
            self.__structures = safe_load(file)

    @property
    def structure_names(self) -> Iterable[str]:
        """The names of all configurations"""
        return self.__structures.keys()

    def create_frame(self, species, structure: str):
        return super().create_frame(species, self.__structures[structure])



# -*- coding: utf-8 -*-
"""This module defines the class :class:`ThermoContribution`, which defines
the building blocks of a :class:`ThermoFrame` function object."""

# stdlib modules
from abc import ABC, abstractmethod
from collections.abc import Sequence, MutableSequence
from typing import Any

# internal modules
from .species import SpeciesDefinition
from .state import InitialState
from ..utilities import Quantity, ParameterDictionary, QuantityDict
from ..utilities.residual import ResidualHandler, ResidualProxy
from ..utilities.types import Map, MutMap


class ThermoContribution(ABC):
    """This abstract class defines the interface of a contribution to a
    thermodynamic state function, as collected in :class:`ThermoFrame` objects.

    A contribution is a reusable part of a thermodynamic model that can be
    recombined meaningfully with other contributions. Examples are standard
    state, ideal mix, ideal gas, Gibbs excess contributions, and
    Helmholtz residual functions (equations of state).

    The definition is based on the ``casadi`` library, and its definition
    is required to build a ``casadi`` evaluation structure as the
    implementation of the belonging equations.

    The usage of this class is mainly indirect by instantiation via the
    :class:`ThermoFactory` objects and parametrisation via the provided parameter
    structures.

    A contribution can overwrite the class attribute ``provides``, helping
    the user to identify feasible contributions if a downstream contribution
    does not find a symbol.
    """

    provides: list[str] = []

    species_definitions: Map[SpeciesDefinition]
    """The map of species definition objects"""

    options: Map[Any]
    """The map of species definition objects"""

    def __init__(self, species: Map[SpeciesDefinition], options):
        self.species_definitions: Map[SpeciesDefinition] = species
        self.options = options
        self.reset()

    def reset(self):
        """Reset the object's state by clearing defined residuals."""
        self.__residuals = ResidualHandler()
        self.__parameters = ParameterDictionary()
        self.__bounds = {}
        self.__vector_props = {}

    @property
    def species(self) -> Sequence[str]:
        """Returns a list of species names"""
        return list(self.species_definitions.keys())

    @abstractmethod
    def define(self, res: MutMap[Quantity]):
        """Abstract method to implement the ``casadi`` expressions
        that make up this contribution.

        See :ref:`standard property names` for a guideline on how to name
        standard properties generated in the contribution implementations.

        :param res: A dictionary with already calculated properties that is to
          be supplemented by the properties calculated in this contribution.
          The values of the dictionaries are of type ``casadi.SX``.
        :param bounds: A dictionary including properties of which the base_unit
          magnitude must stay positive. Solvers can use this information to
          stay within the mathematical domain of the model.
          By convention, if the property is also a result, the same name shall
          be used, and it is not a problem if prior entries are over-written.
          For instance multiple contributions will not allow negative ``T``,
          and all of them shall declare this, as they cannot rely on the others
          being used in the same model.
        :param par: A dictionary with parameters for this contribution. This
          dictionary can be nested. All values are scalar symbols of type
          ``casadi.SX``.

        .. todo::

            - refer to dedicated section the standard property names

        """

    def initial_state(self, state: InitialState, properties: Map[Quantity]) \
            -> MutableSequence[float] | None:
        """When the :class:`ThermoFrame` object is queried for an initial state
        representation and deviates from Gibbs coordinates, The uppermost
        contribution that implements this method and does not return ``None``
        takes the responsibility of calculating that state.

        Hence, normally only Helmholtz models need to implement this method.
        The true model coordinates can however be entirely unconventionally,
        such that it is solely up to the contributions, how to obtain the
        initial state.

        :param state: The default state
        :param properties: The property structure, mapping strings to floats
          or list of floats.
        :return: The initial state or ``None``

        .. seealso:: :meth:`ThermoFrame.initial_state`
        """
        ...

    def add_residual(self, name: str, residual: Quantity,
                     tol_unit: str, tol: float = 1e-7):
        """Define a residual that represents an implicit constraint in the
        thermodynamic model itself. Typical examples are equilibrium
        constraints on apparent species systems and any implicit thermodynamic
        models."""
        self.__residuals.add(name, residual, tol_unit,  tol)

    def add_bound(self, name: str, bound: Quantity,
                  keys: Sequence[str] = None):
        """Add a domain bound to the contribution. This is a property that
        is required to be truly positive.

        If ``bound`` is a vectorial property, its keys must first be registered
        under given ``name`` via :meth:`declare_vector_keys`.

        :param name: Name of the bounded variable
        :param bound: A Scalar or vectorial quantity to remain truly positive
        :param keys: If the quantity is vectorial or ``keys`` are given, they
           will be used to identify the individual element(s). If ``keys`` are
           not given, the vector quantity will use any previously defined
           :meth:`declare_vector_keys` definition instead. If that entry also
           does not exist, species names are assumed to be keys.
        """
        if bound.magnitude.size()[0] == 1 and keys is None:
            self.__bounds[name] = bound
        else:
            if keys is None:
                keys = self.__vector_props.get(name, self.species)
            self.__bounds[name] = QuantityDict.from_vector_quantity(bound, keys)

    @property
    def par_scalar(self):
        """Shortcut method for ``self.parameters.register_scalar``

        .. seealso::
            :class:`~simu.core.utilities.qstructures.ParameterDictionary`
        """
        return self.__parameters.register_scalar

    @property
    def par_vector(self):
        """Shortcut method for ``self.parameters.register_vector``

        .. seealso::
            :class:`~simu.core.utilities.qstructures.ParameterDictionary`
        """
        return self.__parameters.register_vector

    @property
    def par_sparse_matrix(self):
        """Shortcut method for ``self.parameters.register_sparse_matrix``

        .. seealso::
            :class:`~simu.core.utilities.qstructures.ParameterDictionary`
        """
        return self.__parameters.register_sparse_matrix

    @property
    def bounds(self) -> Map[Quantity]:
        return self.__bounds

    @property
    def parameters(self) -> ParameterDictionary:
        return self.__parameters

    @property
    def residuals(self) -> ResidualProxy:
        """Return the defined residuals of this contribution"""
        return self.__residuals

    @property
    def vectors(self) -> Map[Sequence[str]]:
        return self.__vector_props

    def declare_vector_keys(self, name: str, keys: Sequence[str] = None):
        """
        Register a property as a vector property with given keys.
        This way it can be handled correctly by the numeric handler.

        :param name: The name of the property
        :param keys: The names of the keys. If ``None`` (default), the species
          names are used.
        """
        self.__vector_props[name] = self.species if keys is None else keys



"""Unit tests related to the model class"""

from pytest import raises, mark

from simu.core.utilities.errors import DataFlowError, DimensionalityError
from .models import *


def test_parameters_update(caplog):
    caplog.set_level(logging.DEBUG)
    with ParameterTestModel.proxy() as proxy:
        proxy.parameters.update("width", 2, "m")
    assert "area = (length*width) m ** 2" in caplog.text
    assert "width" in proxy.parameters.free
    assert "length" in proxy.parameters.free


def test_parameters_provide():
    width = SymbolQuantity("width", "m")
    with ParameterTestModel.proxy() as proxy:
        proxy.parameters.provide(width=width)
    assert "width" not in proxy.parameters.free
    assert "length" in proxy.parameters.free


def test_parameter_error():
    with raises(KeyError) as err:
        with ParameterTestModel.proxy("Peter") as proxy:
            proxy.parameters.update("Hansi", 2, "m")
    assert "Parameter 'Hansi' not defined in 'Peter'" in str(err)


def test_parameter_missing():
    with raises(DataFlowError) as err:
        with ParameterTestModel.proxy("Peter"):
            pass
    assert "Model 'Peter' has unresolved parameters: 'width'" in str(err.value)


def test_parameter_double():
    width = SymbolQuantity("width", "m")
    with ParameterTestModel.proxy("Peter") as proxy:
        proxy.parameters.provide(width=width)
        with raises(KeyError) as err:
            proxy.parameters.update("width", 2, "m")
    assert "Parameter 'width' already provided in 'Peter'" in str(err.value)


def test_parameter_incompatible():
    temperature = SymbolQuantity("temperature", "K")
    with ParameterTestModel.proxy("Peter") as proxy:
        with raises(DimensionalityError):
            proxy.parameters.provide(width=temperature)
        with raises(DimensionalityError):
            proxy.parameters.update("width", 100, "K")
        proxy.parameters.update("width", 100, "cm")


def test_parameters_access():
    with PropertyTestModel.proxy() as proxy:
        pass
    area = proxy.properties["area"]
    assert f"{area:~}" == "sq(length) m ** 2"


def test_parameters_access_too_early():
    with PropertyTestModel.proxy() as proxy:
        with raises(DataFlowError):
            _ = proxy.properties["area"]


def test_parameters_dont_define():
    model = PropertyTestModel()
    model.define = lambda: None
    with raises(DataFlowError):
        with model.create_proxy():
            pass


def test_parameters_define_other():
    model = PropertyTestModel()

    def my_def():
        """replace define"""
        model.properties["area"] = model.parameters["length"]

    model.define = my_def
    with raises(DimensionalityError):
        with model.create_proxy():
            pass


def test_hierarchy():
    """Test evaluating a simple hierarchical model with just parameters and
    properties"""
    proxy = HierarchyTestModel.top()
    volume = proxy.properties["volume"]
    assert f"{volume:~}" == "(sq(length)*depth) cm * m ** 2"


def test_hierarchy2():
    """Test evaluating a simple hierarchical model with just parameters and
    properties"""
    proxy = HierarchyTestModel2.top()
    volume = proxy.properties["volume"]
    assert f"{volume:~}" == "(depth*sq((2*radius))) cm ** 3"

@mark.parametrize("simple_material_definition", [["H2O", "NO2"]], indirect=True)
def test_material(simple_material_definition, material_test_model):
    material = simple_material_definition.create_flow()

    with material_test_model().create_proxy() as model:
        assert "inlet" in model.materials
        model.materials.connect("inlet", material)


@mark.parametrize("simple_material_definition", [["KMnO4"]], indirect=True)
def test_wrong_material(simple_material_definition, material_test_model):
    material = simple_material_definition.create_flow()
    with raises(ValueError) as err:
        with material_test_model().create_proxy() as model:
            model.materials.connect("inlet", material)
    assert "incompatible" in str(err)


@mark.parametrize("simple_material_definition", [["H2O", "NO2"]], indirect=True)
def test_material_reuse_def(simple_material_definition):
    material = simple_material_definition.create_flow()

    with MaterialTestModel2().create_proxy() as model:
        assert "inlet" in model.materials
        model.materials.connect("inlet", material)


def test_residual_def():
    _ = ResidualTestModel().top()


def test_residual():
    with ResidualTestModel2().create_proxy() as model:
        pass
    res = model.residuals["Hubert"]
    assert f"{res.value:~}" == "Hubert K"



from typing import Type
from collections.abc import Mapping, Collection

from .state import StateDefinition
from .species import SpeciesDefinition
from .frame import ThermoFrame
from .contribution import ThermoContribution


class ThermoFactory:
    """The ``ThermoFactory`` class hosts the definitions for the *model
    contributions*, enabling it to create instances of thermodynamic models of
    class :class:`ThermoFrame`.

    The class is largely meant to be a singleton, but to keep doors open,
    static attributes are avoided."""

    def __init__(self):
        """Parameter-less constructor, initialising the data structure
        to host contribution definitions"""
        self.__contributions = {}
        self.__state_definitions = {}

    def register_state_definition(self, definition_cls: Type[StateDefinition]):
        """Register a new state definition with the name of its class.

        :param definition_cls: The state definition to register
        """
        name = definition_cls.__name__
        if name in self.__state_definitions:
            raise ValueError(f"State definition '{name}' already defined.")
        self.__state_definitions[name] = definition_cls

    def register(self, *contributions: Type[ThermoContribution]):
        """Registers contributions under the names of their classes.

        The contributions must be concrete subclasses of
        :class:`ThermoContribution`, and their names must be unique.

        :param contributions: The contributions to register, being classes
          (not instances)
        """
        for class_ in contributions:
            name = class_.__name__
            if name in self.__contributions:
                raise ValueError(f"Contribution '{name}' already defined.")
            self.__contributions[name] = class_

    @property
    def contribution_names(self) -> Collection[str]:
        """This property contains the full names of all registered
        contributions"""
        return set(self.__contributions.keys())

    def create_frame(self, species: Mapping[str, SpeciesDefinition],
                     configuration: Mapping) -> ThermoFrame:
        """This factory method creates a :class:`ThermoFrame` object from the
        given ``configuration``, and is the recommended way to create
        :class:`ThermoFrame` objects.

        :param species: A dictionary mapping names to species definitions
        :param configuration: A nested dictionary with the following root
          entries:

            - ``state``: A string identifier that represents the type of state
              as defined by :meth:`register_state_definition`.
            - ``contributions``: A list of strings, representing the names
              of the contributions to stack. These identifiers must have been
              defined upfront by calls to :meth:`register`.
              A contribution entry can also be a dictionary with the following
              keys:

                - ``cls``: The contribution class (required). If this is the
                  only key defined, its effect is as if the sole string of the
                  class name was provided.
                - ``name``: The name of the contribution in the created
                  :class:`ThermoFrame` object. This name must be unique within
                  the frame definition. It will also be used to define the
                  contribution parameter structure.
                  If skipped, the name will be the same as ``cls``. To be
                  unique, this doesn't work if one contribution class is used
                  multiple times.
                - ``options``: Any data structure that is accepted (and
                  hopefully documented) for the particular contribution.
                  If skipped, an empty dictionary is used.

        :return: The thermodynamic model (:class:`ThermoFrame`) object
        """
        contributions = {}
        for item in configuration["contributions"]:
            if isinstance(item, dict):
                class_ = item["cls"]
                name = item.get("name", class_)
                options = item.get("options", {})
            else:
                name, class_, options = item, item, {}
            class_ = self.__contributions[class_]
            if name in contributions:
                raise ValueError(f"Duplicate contribution name '{name}'")
            contributions[name] = class_, options

        state_def_cls = self.__state_definitions[configuration["state"]]
        result = ThermoFrame(species, state_def_cls(), contributions)

        # set default state
        default = configuration.get("default_state", None)
        if default is not None:
            # make sure the values are float
            default = [
                float(default[0]),
                float(default[1]),
                list(map(float, default[2]))
            ]
        result.default = default
        return result



from abc import ABC
from .base import Model


class AModel(Model, ABC):
    """This is a wrapper class to enable short notations for common method
    calls, and by that allowing to shorten the syntax when creating models,
    while still keeping the underlying structure.

    The abbreviations are given in the following table:

    .. currentmodule:: simu.core.model

    ============ ==============================================================
    Abbreviation ... resolves to
    ============ ==============================================================
    pa           :attr:`parameters <simu.Model.parameters>`
    pad          :meth:`parameters.define <parameter.ParameterHandler.define>`
    pas          :meth:`parameters.static <parameter.ParameterHandler.static>`
    pr           :attr:`properties <simu.Model.properties>`
    prd          :meth:`properties.declare <property.PropertyHandler.declare>`
    h            :attr:`hierarchy <simu.Model.hierarchy>`
    hd           :meth:`hierarchy.declare <hierarchy.HierarchyHandler.declare>`
    ha           :meth:`hierarchy.add <hierarchy.HierarchyHandler.add>`
    m            :attr:`materials <simu.Model.materials>`
    md           :meth:`materials.define_port <material.MaterialHandler.define_port>`
    mcf          :meth:`materials.create_flow <material.MaterialHandler.create_flow>`
    mcs          :meth:`materials.create_state <material.MaterialHandler.create_state>`
    r            :attr:`residuals <simu.Model.residuals>`
    ra           :meth:`residuals.add <simu.core.utilities.residual.ResidualHandler.add>`
    ============ ==============================================================
    """

    # parameter handler methods
    @property
    def pa(self):
        return self.parameters

    @property
    def pad(self):
        return self.parameters.define

    @property
    def pas(self):
        return self.parameters.static

    # property handler methods
    @property
    def pr(self):
        return self.properties

    @property
    def prd(self):
        return self.properties.declare

    # hierarchy handler methods
    @property
    def h(self):
        return self.hierarchy

    @property
    def hd(self):
        return self.hierarchy.declare

    def ha(self):
        return self.hierarchy.add

    # materials handler methods
    @property
    def m(self):
        return self.materials

    @property
    def md(self):
        return self.materials.define_port

    @property
    def mcf(self):
        return self.materials.create_flow

    @property
    def mcs(self):
        return self.materials.create_state

    # residuals handler methods
    @property
    def r(self):
        return self.residuals

    @property
    def ra(self):
        return self.residuals.add





from symtable import Function
from typing import Callable, Sequence, Any
from copy import deepcopy
from dataclasses import dataclass, field
from time import time
import sys
from io import TextIOBase

from casadi import MX, jacobian, jtimes, Function
from numpy import array, argmin, argmax, abs, squeeze, log10
from scipy.sparse import csc_array

try:  # use pypardiso if installed
    from pypardiso import spsolve
except ImportError:  # use scipy if not
    from scipy.sparse.linalg import spsolve

from ..model.numeric import NumericHandler
from ..utilities import Quantity, QFunction
from ..utilities.output import ProgressTableOutput
from ..utilities.types import Map, NestedMutMap, NestedMap
from ..utilities.configurable import Configurable
from ..utilities.errors import IterativeProcessInterrupted, NonSquareSystem

_VEC, _STATE = NumericHandler.VECTORS, NumericHandler.STATE_VEC
_RES, _BOUND = NumericHandler.RES_VEC, NumericHandler.BOUND_VEC


@dataclass
class SimulationSolverIterationReport:
    """This data class object is provided for each iteration during a
    :class:`SimulationSolver` run.
    """
    max_err: float
    r"""For each :class:`~simu.core.utilities.residual.Residual`, the 
    quotient of residual value :math:`r_i` and tolerance :math:`t_i` is 
    calculated. ``max_err`` is the maximum absolute value of these quotients:
    
    .. math:: \mathrm{MET} = \max_i \frac{r_i}{t_i}
    """

    max_res_name: str
    """The name of the :class:`~simu.core.utilities.residual.Residual` which
    causes the value of :attr:`max_err`"""

    relax_factor: float
    """The applied relaxation factor to stay within the domain of the process
    model and the thermodynamic models, according to the defined bounds."""

    min_alpha_name: str
    """The name of the bound that is most limiting and therefore causing the
    value of :attr:`relax_factor`"""

    duration: float
    """The accumulative duration of the solving process inclusive the given
    iteration"""

    lmet: float = field(init=False)  # logarithmic max error to tolerance
    r"""The logarithmic (base 10) value of :attr:`max_err` :math:`r_i / t_i`,
    practically defined as
    
    .. math::
        
        \mathrm{LMET} = \log_{10} \left (
            \max_i \frac{r_i}{t_i} + 10^{-8} \right )

    The offset is introduced to not cause ``NaN`` values for the lucky case in
    which all residuals are exactly zero. This can however easily happen for
    linear systems. Anyhow, :math:`\mathrm{LMET} < 1` is already a sufficient
    condition for convergence.    
    """

    def __post_init__(self):
        self.lmet = log10(self.max_err + 1e-8)


@dataclass
class SimulationSolverReport:
    """The data class object returned from a :class:`SimulationSolver` run
    """
    iterations: Sequence[SimulationSolverIterationReport]
    """A :class:'SimulationSolverIterationReport` object for each performed
    iteration"""

    final_state: Sequence[float]
    """The numerical final state of the model.
    
    .. important::
    
        This state is not suitable for handling initial values in a robust way,
        as it can be very sensitive to minor model changes that for instance
        impact the liquid volumes of equations of state.
        
        Instead, use :meth:`simu.NumericHandler.export_state`,
        :meth:`~simu.NumericHandler.import_state` and 
        :meth:`~simu.NumericHandler.retain_state`.
    """

    prop_func: Callable[[Sequence[float]], NestedMap[Quantity]]
    """The function to calculate all properties of the model as function of
    the state."""

    @property
    def properties(self) -> NestedMap[Quantity]:
        """This property returns all properties of the model, evaluated on
        the :attr:`final_state` attribute. For larger models, this causes
        noticeable computational effort. For this reason, this evaluation is
        only done on demand."""
        return self.prop_func(self.final_state)


SimulationSolverCallback = Callable[
    [int,
     SimulationSolverIterationReport,
     Sequence[float],
     Callable[
         [Sequence[float]],
         NestedMap[Quantity]]
     ],
     bool]
"""A function (type) to act as a call-back in the :class:`SimulationSolver`
solving process. The arguments are as follows:

- ``iteration``: The iteration number as integer, incrementing from zero
- ``report``: The (:class:`SimulationSolverIterationReport`) object
- ``state``: The internal state of the model at given iteration as a sequence of
  floats
- ``prop_func``: A function to calculate all model properties for the given
  state.

The last argument is provided instead of the property structure itself, as it
might be expensive to calculate the entire property structure in each iteration.
This way it can be done on demand.

The callback shall return ``True``, if the solver is to continue, or ``False``
otherwise. In the latter case, the solver will raise a 
:class:`~simu.core.utilities.errors.IterativeProcessInterrupted` exception.

Example:

.. code-block::
   :linenos:

    from pprint import pprint
    
    def my_callback(iteration, report, state, prop_func):
        # This can be a lot to print
        all_properties = prop_func(state)
        pprint(all_properties)

"""

class SimulationSolver(Configurable):
    r"""
    The simulation solver assumes both thermodynamic and model parameters to
    be constant, aiming to find the state variable values such that all
    residuals evaluate to zero within their tolerance.
    """
    # noinspection PyUnusedLocal
    # Options are parsed via inspection
    def __init__(self, model: NumericHandler, *,
                 max_iter: int = 30,
                 gamma: float = 0.9,
                 wall: float = 1e-20,
                 output: TextIOBase|str = "stdout",
                 call_back_iter: SimulationSolverCallback = None,
                 retain_solution: bool = True):
        r"""On construction, the solver object requires a
        :class:`~simu.NumericHandler` object. The solver object can then be
        reused for multiple solver runs, for instance with variable parameter
        values (sensitivity study).

        :param model: The numeric handler of a model, in most cases obtained by
          the expression ``NumericHandler(ModelClass.top())``.
        :param max_iter: The maximum number of iterations (default 30).

          .. note::

            Normally, 30 iterations should be sufficient. In other words, if the
            model is not converged after 30 iterations, chances are quite low
            that it still will converge at all. The advice would be to try to
            improve the starting values and to investigate whether the model is
            properly posed.

        :param gamma: As described above, :math:`\gamma` (default 0.9) is the
          fraction of the step-length applied by the solver before hitting the
          domain boundary. Normally, changing the value is not required.
          Generally, a lower value makes the model more robust against
          non-linear domain boundaries (and thus linearisation errors causing
          the state to exit the domain. A higher value yields slightly faster
          convergence, if the solution is in comparison with the initial values
          very close to the domain boundary.
        :param wall: Either if there is no solution within the domain of the
          model (for instance: The material balance forces some of the species
          flows in a stream to be negative), or if the solver for other reasons
          is forced to try to leave the model domain, the state will move closer
          and closer to the domain boundary and not revert. At some point,
          :math:`\gamma` becomes ridiculously small, and we need to give up.
          This threshold value is defined by ``wall`` (default ``1e-20``).
        :param output: The io-stream to direct the solver output to, or a
          descriptive string (case-insensitive):

          - ``"stdout"``: The output will be written to standard out (default)
          - ``"none"``: No output will be printed.

          .. note::

            Instead of printing, one might either analyse the returned
            :class:`~simu.core.solver.simulation.SimulationSolverReport` project
            after the run, or utilise the ``call_back_iter`` callback and
            process the iteration progress from there.

        :param call_back_iter: A callback function (default ``None``),
          see :data:`~simu.core.solver.simulation.SimulationSolverCallback`,
          to intercept the solving process. The returned boolean variable
          determines whether the solver iteration is continued or not.

        :param retain_solution: Whether the solver shall, on success, retain the
          obtained state in the model, such that it can be exported via
          :meth:`~simu.NumericHandler.export_state` and be reused as the initial
          values for the next solving process.
        """
        super().__init__(exclude=["model"])
        self._model = model

        # store arguments (parameters) so the user can change them
        args = deepcopy(model.arguments)
        # store size of state
        self.__state_size = args[_VEC][_STATE].magnitude.size()[0]
        res_size = len(model.vector_res_names(NumericHandler.RES_VEC))

        if self.__state_size != res_size:
            raise NonSquareSystem(self.__state_size, res_size)

        # user shall not think that putting a state here has any effect
        del args[_VEC][_STATE]
        self.__model_parameters : NestedMutMap[Quantity] = args

    def solve(self, **kwargs: Any) -> SimulationSolverReport:
        """
        This method triggers iterative the solving process. This takes less than
        0.1 seconds for small models, and increases with model complexity.
        For large models, the time per iteration is due to the solving of linear
        systems cubic in system size, though the model structure might render
        this a conservative estimate.

        With `pypardiso`_ installed, the solving of the linear systems is
        performed on all available CPU cores.

        For each given keyword argument, the
        :meth:`~simu.core.utilities.configurable.Configurable.set_option`
        method is invoked, giving the same effect as if the option was provided
        with the constructor.

        :return: The report including the iteration sequence
        """
        for name, value in kwargs.items():
            self.set_option(name, value)
        opt = self.options
        model = self._model
        start_time = time()
        residual_names = model.vector_res_names(_RES)
        bound_names = model.vector_res_names(_BOUND)
        reports = []

        output = self.__find_output()

        table = ProgressTableOutput({
            "lmet": ("LMET", "{:5.1f}"),
            "relax_factor": ("Alpha", "{:7.2g}"),
            "duration": ("Time", "{:6.1g}"),
            "min_alpha_name": ("Limit on bound", "{:>40s}"),
            "max_res_name": ("Max residual", "{:>40s}")
        }, row_dig=5, row_head="Iter", stream=output)

        funcs = self._prepare_functions()
        x = self.initial_state

        for iteration in range(opt["max_iter"]):
            # evaluate system (matrix and rhs)
            r, dr_dx = funcs["f_r"](x)
            r = squeeze(array(r))
            dr_dx = csc_array(dr_dx)

            # assess error
            max_err_idx = argmax(abs(r))
            max_res_name = residual_names[max_err_idx]
            max_err = abs(r[max_err_idx])
            if max_err < 1:
                break

            # calculate full update
            dx = -spsolve(dr_dx, r)

            # find relaxation factor
            a = squeeze(array(funcs["f_b"](x, dx)))
            a = a[0 < a]
            alpha, min_alpha_name = 1, ""
            if len(a):
                min_a_idx = int(argmin(a))
                if a[min_a_idx] * opt["gamma"] < 1:
                    alpha = a[min_a_idx] * opt["gamma"]
                    min_alpha_name = bound_names[min_a_idx]
                if alpha < opt["wall"]:
                    msg = f"Relaxation factor is below {opt["wall"]}, " \
                          "no solution found"
                    raise ValueError(msg)
            # apply update
            x = x + alpha * dx

            # reporting
            duration = time() - start_time
            reports.append(SimulationSolverIterationReport(
                max_err=float(max_err),
                max_res_name=max_res_name,
                relax_factor=float(alpha),
                min_alpha_name=min_alpha_name,
                duration=duration
            ))
            if opt["call_back_iter"] is not None:
                cb_result = opt["call_back_iter"](
                    iteration, reports[-1], x.magnitude,
                    lambda x_arg: funcs["f_y"]({"x": Quantity(x_arg)})
                )
                if not cb_result:
                    msg = "Solver iterations interrupted by callback"
                    raise IterativeProcessInterrupted(msg)
            table.row(reports[-1], iteration)
        else:
            msg = f"Model did not converge after {opt["max_iter"]} iterations"
            raise ValueError(msg)

        # reporting
        duration = time() - start_time
        reports.append(SimulationSolverIterationReport(
            max_err=float(max_err),
            max_res_name=max_res_name,
            relax_factor=1,
            min_alpha_name="",
            duration=duration
        ))
        table.row(reports[-1], iteration)

        # retain state if desired
        if opt["retain_solution"]:
            model.retain_state(x, self.model_parameters["thermo_params"])

        return SimulationSolverReport(
            iterations=reports,
            final_state=x,
            prop_func=lambda z: funcs["f_y"]({"x": Quantity(z)})
        )

    def __find_output(self) -> TextIOBase|None:
        output = self.options["output"]
        if isinstance(output, str):
            opts = {"stdout": sys.stdout, "none": None}
            try:
                return opts[output.lower()]
            except KeyError:
                raise ValueError(f"Invalid stream name '{output}'")
        return output

    def _prepare_functions(self) -> Map[Callable]:
        # prepare
        #  - a casadi MX function x -> (r, dr/dx)
        #  - a casadi MX function: (x, dx) -> (a_i = b_i / (db_i/dx_j) * dx_j)
        # prepare a QFunction x -> (y_m, y_t)
        param = deepcopy(self.__model_parameters)

        param[_VEC][_STATE] = (Quantity(x := MX.sym("x", self.__state_size)))
        res = self._model.function(param, squeeze_results=False)
        r, b = res[_VEC][_RES], res[_VEC][_BOUND]
        dx = Quantity(MX.sym("x", self.__state_size))
        return {
            "f_r": Function("f_r", [x], [r, jacobian(r, x)]),
            "f_b": Function("f_b", [x, dx], [-b / jtimes(b, x, dx)]),
            "f_y": QFunction({"x": Quantity(x)}, res)
        }

    @property
    def initial_state(self):
        """Freshly extract the initial values from the model. These might have
        been changed after the solver class was instantiated"""
        args = self._model.arguments
        return args[NumericHandler.VECTORS][NumericHandler.STATE_VEC]

    @property
    def model_parameters(self) -> NestedMutMap[Quantity]:
        """A convenience property to access the parameters of the model as a
        mutable object. The state variables are removed in this instance, as
        these are rather provided by the solver during the iterative solving
        process."""
        return self.__model_parameters

    @property
    def _arg_validations(self):
        between = Configurable._validate_between
        return {
            "max_iter": between(1, 10000),
            "gamma": between(0.1, 0.999),
            "call_back_iter": {
                "f": lambda x: x is None or callable(x),
                "msg": "must be callable",
            },
            "output": {
                "f": lambda x: isinstance(x, str)  or isinstance(x, TextIOBase),
                "msg": "must be a stream or a qualified string"
            }
        }



from numpy import linspace, log, min
from matplotlib import pyplot

parameters = {
    "T_0": 298.15,  # K
    "white": {
        "DH": 0,  # J/mol
        "S0": 51.39,  # J/(mol.K)  # adapted to transition temperature
        "CP": 26.99  # J/(mol.K)
    },
    "grey": {
        "DH": -2090,  # J/mol
        "S0": 44.14,  # J/(mol.K)
        "CP": 25.77  # J/(mol.K)
    }
}


def mu(temperature, species):
    dh = parameters[species]["DH"]
    s0 = parameters[species]["S0"]
    cp = parameters[species]["CP"]
    t_0 = parameters["T_0"]
    return dh - temperature * s0 \
        + cp * (temperature - t_0 - log(temperature / t_0))

def main():
    T = linspace(278.15, 293.15)
    pyplot.figure(figsize=(8, 4))
    pyplot.plot(T - 273.15, min([mu(T, "white"), mu(T, "grey")], axis=0) / 1000,
                "-", color="#dddddd", lw=10, label="stable phase")
    pyplot.plot(T - 273.15, mu(T, "white") / 1000, "k-", label="white tinn")
    pyplot.plot(T - 273.15, mu(T, "grey") / 1000, "b-", label="grey tinn")
    pyplot.plot([13.2, 13.2], [-15.2, -14.8], "r:", label="Transition temperature")
    pyplot.grid()
    pyplot.xlim([5, 20])
    pyplot.ylim([-15.2, -14.8])
    pyplot.legend()
    pyplot.xlabel("Temperature [$^\\circ$C]")
    pyplot.ylabel("Chemical potential [kJ/(mol$\\,$K)]")
    pyplot.show()


if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-

# internal modules
from simu import ThermoContribution, exp, conditional, sqrt, qpow, qsum
from simu.core.utilities import SymbolQuantity

from .rk import (RedlichKwongEOSLiquid, RedlichKwongEOSGas,
                 RedlichKwongAFunction, RedlichKwongBFunction,
                 RedlichKwongMFactor)


class LinearMixingRule(ThermoContribution):
    r"""This linear mixing rule represents any contribution that computes a
    lumped quantity as weighted sum over the molar quantities:

    .. math::

        x = \sum_i x_i\,n_i

    The contribution requires an ``option`` dictionary with the following
    entries:

      - ``target``: The name of the property :math:`x`.
      - ``source``: The name of the property :math:`x_i`. If omitted, it
        will be generated as the target name with ``_i`` appended.
    """

    def define(self, res):
        target = self.options["target"]
        source = self.options.get("source", target + "_i")
        res[target] = res[source].T @ res["n"]


class NonSymmetricMixingRule(ThermoContribution):
    r"""The mixing rule combines the pure species a-contributions ``a_i``
    (:math:`a_i`) into a lumped ``a`` property.

    The contribution requires an ``option`` dictionary with the following
    entries:

      - ``target``: The name of the property :math:`a`.
      - ``source``: The name of the property :math:`a_i`. If omitted, it
        will be generated as the target name with ``_i`` appended.

    Implemented with sparse binary interaction coefficient structure, the
    contribution includes a symmetric interaction (:math:`k`) and an
    antisymmetric contribution (:math:`l`).

    .. math::

        a = \sum_{ij} \sqrt{a_i(T)\,a_j(T)} \left [
              n_i\,n_j - 2\,n_i\,n_j\,k_{ij}(T) -
              \frac2N\,(n_i^2\,n_j - n_i\,n_j^2)\,l_{ij}(T)
            \right ]

    These dimensionless interaction coefficients can be a function of
    temperature according to

    .. math::

        k_{ij}(T) &:= k_{0,ij} +
                      k_{1,ij}\,\left (1 - \frac{T}{T_\mathrm{ref}} \right ) +
                      k_{2,ij}\,\left (1 - \frac{T_\mathrm{ref}}{T} \right )\\
        l_{ij}(T) &:= l_{0,ij} +
                     l_{1,ij}\,\left (1 - \frac{T}{T_\mathrm{ref}} \right ) +
                     l_{2,ij}\,\left (1 - \frac{T_\mathrm{ref}}{T} \right )

    Each pair of species can only be defined once. The pre-factor 2 is used to
    yield the same parameter values as if :math:`k` was a symmetric matrix and
    :math:`l` was antisymmetric.

    Despite what above equation suggests based on the double sum, the
    complexity of this contribution in terms of both memory and runtime is
    linear in the number of species and linear in the number of non-zero
    interaction parameters. This is achieved using the relationship

    .. math::
        \sum_{ij} \sqrt{a_i(T)\,a_j(T)} n_i\,n_j
          = \left (\sum_i \sqrt{a_i(T)}\,n_i \right )^2
    """

    def define(self, res):
        target = self.options["target"]

        def extract():
            source = self.options.get("source", target + "_i")

            temp, n, a_i = res["T"], res["n"], res[source]
            tau = temp / self.par_scalar("T_ref", "K")
            tau_1, tau_2 = 1 - tau, 1 - 1 / tau
            a_n = sqrt(a_i) * n
            N = qsum(n)
            return n, N, a_n, tau_1, tau_2

        n, N, a_n, tau_1, tau_2 = extract()

        # pre-factors can be reused to minimise graph size
        cache = [{}, {}]

        def c_1(i: int, j: int) -> SymbolQuantity:
            if (i, j) not in cache[0]:
                idx_i = self.species.index(i)
                idx_j = self.species.index(j)
                cache[0][(i, j)] = a_n[idx_i] * a_n[idx_j]
            return cache[0][(i, j)]

        def c_2(i: int, j: int) -> SymbolQuantity:
            if (i, j) not in cache[1]:
                idx_i = self.species.index(i)
                idx_j = self.species.index(j)
                cache[1][(i, j)] = c_1(i, j) * (n[idx_i] - n[idx_j])
            return cache[1][(i, j)]

        def symmetric():
            contributions = []
            for name, factor in [("k_1", 1), ("k_2", tau_1), ("k_3", tau_2)]:
                try:
                    pairs = self.options[name]
                except KeyError:
                    continue
                coeff = self.par_sparse_matrix(name, pairs, "dimless")
                if coeff:
                    term = sum(c_1(i, j) * c for i, j, c in coeff.pair_items())
                    contributions.append(factor * term)
            return -2 * sum(contributions) if contributions else None

        def asymmetric():
            contributions = []
            for name, factor in [("l_1", 1), ("l_2", tau_1), ("l_3", tau_2)]:
                try:
                    pairs = self.options[name]
                except KeyError:
                    continue
                coeff = self.par_sparse_matrix(name, pairs, "dimless")
                if coeff:
                    term = sum(c_2(i, j) * c for i, j, c in coeff.pair_items())
                    contributions.append(factor * term)
            return -2 / N * sum(contributions) if contributions else None

        sym = symmetric()
        asym = asymmetric()
        res[target] = qsum(a_n) ** 2
        if sym is not None:
            res[target] += sym
        if asym is not None:
            res[target] += asym

        self.add_bound("T", res["T"])


class CriticalParameters(ThermoContribution):
    r"""This class does not perform any calculations, but provides the basic
    critical parameters as a basis for the typical equation of state
    contributions.

    The following parameters need to be provided (all as species vector):

    ======== ============== =========================
    Property Symbol         Description
    ======== ============== =========================
    T_c      :math:`T_c`    Critical temperatures [K]
    p_c      :math:`p_c`    Critical pressure[Pa]
    omega    :math:`\omega` Acentric factor [-]
    ======== ============== =========================

    The same symbols will just be published as intermediate results for the
    actual model contributions to be consumed.
    """

    provides = ["_T_c", "_p_c", "_omega"]

    def define(self, res):
        res["_T_c"] = self.par_vector("T_c", self.species, "K")
        res["_p_c"] = self.par_vector("p_c", self.species, "bar")
        res["_omega"] = self.par_vector("omega", self.species, "dimless")


class VolumeShift(ThermoContribution):
    r"""This class does not perform any calculations, but provides volume
    shift parameters to be used via mixing rules as the C-parameter in
    equations of state. The following parameter needs to be provided as a
    species vector:

    ======== ============== =================================
    Property Symbol         Description
    ======== ============== =================================
    c_i      :math:`c_i`    Volume shift parameter [m**3/mol]
    ======== ============== =================================
    """

    provides = ["_ceos_c_i"]

    def define(self, res):
        res["_ceos_c_i"] = self.par_vector("c_i", self.species, "m**3/mol")


class BostonMathiasAlphaFunction(ThermoContribution):
    r"""This contribution represents the Mathias alpha function with the
    Boston-Mathias extrapolation for supercritical temperatures.

    The following properties need to be provided upstream:

    ======== ============== ===========================================
    Property Symbol         Description
    ======== ============== ===========================================
    T        :math:`T`      Actual temperatures [K]
    T_c      :math:`T_c`    Critical temperatures [K]
    m_factor :math:`m_i`    m-factor as function of acentric factor [-]
    ======== ============== ===========================================

    Additionally, the contribution requires a polar parameter :math:`\eta_i`,
    named ``eta``. We define the root of the reduced temperature as
    :math:`\tau_i := \sqrt{T/T_{c,i}}`. Then, we define for
    :math:`\tau_i \le 1`:

    .. math::

        \alpha_i^{\frac12} =
          1 + m_i\,(1 - \tau_i) - \eta_i\,(1 - \tau_i)(0.7 - \tau_i^2)

    .. important::

        While the paper :cite:p:`Mathias1983` is currently unavailable to me, I
        must recognize that all but :cite:p:`AspenTech2001` quote the expression
        differently, namely:

        .. math::

            \alpha_i^{\frac12} =
              1 + m_i\,(1 - \tau_i) - \eta_i\,
              (1 - \boxed{\tau_i^2})(0.7 - \tau_i^2)

        Likely, there was a misprint in the original paper
        :cite:p:`Mathias1983`, as the parameter :math:`\eta_i` shall not alter
        the impact of the acentric factor :math:`\omega_i` at
        :math:`T = 0.7\, T_{c,i}`, as this point is defined to match the
        saturation pressure via

        .. math::

            \omega_i = -\log_{10}
              \frac{p^\mathrm{sat}_i(0.7\cdot T_{c,i})}{p_{c, i}}

        The reason why AspenTech could and did use the correct form is that the
        author P. Mathias was working for the company at the time of these
        developments, and the co-workers were not dependent on the journal
        publication.

    As described in Appendix (:ref:`alpha_extensions`), the extrapolation into
    the super-critical region is implemented as

    .. math::

        \alpha_i^{\frac12} = \left [\frac{c}{d}(1-\tau^{d})\right ]
        \quad\text{with}\quad c = m + 0.3\eta
        \quad\text{and}\quad d = 1 + \frac{4\,\eta}{c} + c

    The calculated vector is provided as a property called ``alpha``
    """

    provides = ["_alpha"]

    def define(self, res):
        eta = self.par_vector("eta", self.species, "dimless")
        temp, critical_temp, m_fac = res["T"], res["_T_c"], res["_m_factor"]
        tau = temp / critical_temp
        stau = sqrt(tau)

        # define sub and super-critical expression
        alpha_sub = 1 + m_fac * (1 - stau) - eta * (1 - stau) * (0.7 - tau)

        bm_c = m_fac + 0.3 * eta
        bm_d = 1 + bm_c + 4 * eta / bm_c
        alpha_sup = exp(bm_c / bm_d * (1 - qpow(stau, bm_d)))

        alpha = conditional(tau > 1, alpha_sub, alpha_sup)
        # result is square of above
        res["_alpha"] = alpha * alpha

        res["T"] = temp


"""This module defines data structures to host the global species list"""

from typing import Self
from dataclasses import dataclass, field
from collections.abc import Mapping, Iterator, Sequence

from ..utilities import Quantity
from ..utilities.structures import Map
from ..utilities.molecules import FormulaParser

_PARSER = FormulaParser()

@dataclass
class SpeciesDefinition:
    """
    This class holds a definition of a species and provides, based on the
    formula, the molecular weight, charge, and a dictionary of element
    composition.

    >>> a = SpeciesDefinition("H3PO4")
    >>> print(f"{a.molecular_weight:~.3f}")
    97.993 g / mol
    >>> a.elements
    {'H': 3, 'P': 1, 'O': 4}
    >>> a = SpeciesDefinition("PO4:3-")
    >>> print(f"{a.charge:~}")
    -3 e / mol
    """
    formula: str
    """The formula as it was given in the constructor. The admitted formula 
    syntax is described with examples for the
    :class:`FormulaParser <simu.core.utilities.molecules.FormulaParser>` class.
    """
    molecular_weight: Quantity = field(init=False)
    """The molecular weight determined by summing up the atomic weights of 
    the contained atoms; quantum effects and electron masses are neglected."""
    charge: Quantity = field(init=False)
    """The electronic charge of the species' molecule"""
    elements: Map[int] = field(init=False)
    """A dictionary, mapping the occurring atoms to their amount in the 
    species' molecule"""

    def __post_init__(self):
        self.elements = dict(_PARSER.parse(self.formula))
        self.molecular_weight = _PARSER.molecular_weight(self.formula)
        self.charge = _PARSER.charge(self.formula)


class SpeciesDB(Mapping[str, SpeciesDefinition]):
    """Based on a dictionary of species names to formulae, this class
    represents a dictionary of the species names to species definitions.

    .. note::
        For now, this class is quite primitive, but might be extended to handle
        more meta-data, such as CAS registry numbers and species aliases.
    """
    def __init__(self, formulae: Mapping[str, str]):
        """Create a species collection based on a mapping of species names
        to their formulae."""
        self.__species = {n: SpeciesDefinition(f) for n, f in formulae.items()}

    def __getitem__(self, key: str) -> SpeciesDefinition:
        return self.__species[key]

    def __len__(self) -> int:
        return len(self.__species)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__species)


    def get_sub_db(self, keys: Sequence[str]) -> Self:
        """Get a subset of species definitions as a new SpeciesDB object."""
        return {n: self[n] for n in keys}



"""This module implements functionality related to parameter handling"""

from typing import Optional
from collections.abc import Mapping, Iterable, Iterator

from ..utilities import Quantity, SymbolQuantity
from ..utilities.types import Map, MutMap
from ..utilities.errors import DataFlowError


class ParameterHandler(Map[Quantity]):
    """This class, being instantiated as the :attr:`simu.Model.parameters`
    attribute, allows to define and access process parameters.

    During :meth:`simu.Model.interface`, the model defines parameters with and
    without pre-defined values.

    Within the ``with`` block, the parent module connects parameters by
    providing external symbols. Here, the unit of measurement must be
    compatible to the parameters' definition.

    During :meth:`simu.Model.define` checks that all required symbols (the ones
    without default values) are connected. It keeps track of the non-connected
    and static parameters, as they will need to be used as arguments for the
    overall model function.
    """

    static_parameters: MutMap[Quantity] = {}
    static_values: MutMap[Quantity] = {}

    def __init__(self, static_id: str):
        self.__params: MutMap[Quantity] = {}
        self.__values: MutMap[Quantity] = {}
        self.__static_used_names: set[str] = set()
        self.__static_id = static_id

    def __len__(self) -> int:
        return len(self.__params)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__params)

    def define(self,
               name: str,
               value: Optional[float] = None,
               unit: Optional[str] = "dimless") -> None:
        """Define a parameter from within the :meth:`simu.Model.interface`
        method."""

        if name in self.__params:
            raise KeyError(f"Parameter '{name}' already defined")

        # all parameters are stored as symbols with unit
        self.__params[name] = SymbolQuantity(name, unit)
        if value is not None:
            # these have default values and don't need to be provided
            self.__values[name] = Quantity(value, unit)

    def static(self,
               name: str,
               value: float,
               unit: Optional[str] = "dimless") -> SymbolQuantity:
        """Define a static parameter over all instances of the associated
        model."""
        if name in self.__static_used_names:
            raise KeyError(f"Static Parameter '{name}' already defined")
        full_name = f"{self.__static_id}/{name}"
        cls = ParameterHandler
        if full_name not in cls.static_parameters:
            result = SymbolQuantity(name, unit)
            cls.static_parameters[full_name] = result
            cls.static_values[full_name] = Quantity(value, unit)
            return result
        else:
            return cls.static_parameters[full_name]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return

    def __getitem__(self, name: str) -> Quantity:
        """Return the symbol of a defined parameter. This is to be used within
        the :meth:`simu.Model.define` method."""
        return self.__params[name]

    def get_static(self, name: str) -> SymbolQuantity:
        """Return a static parameter of given name"""
        cls = ParameterHandler
        full_name = f"{self.__static_id}/{name}"
        return cls.static_parameters[full_name]

    def static_names(self) -> Iterable[str]:
        """An iterator over all names of defined static parameters"""
        params = ParameterHandler.static_parameters
        prefix = f"{self.__static_id}/"
        return filter(lambda key: key.startswith(prefix), params.keys())

    def create_proxy(self) -> "ParameterProxy":
        """Create a proxy object for configuration in hierarchy context"""
        return ParameterProxy(self, self.__params, self.__values)


class ParameterProxy(Mapping[str, Quantity]):
    """This class is instantiated by the parent's
    :class:`~simu.core.model.parameter.ParameterHandler` to configure the
    parameter connections from the parent context."""

    def __init__(self, handler: ParameterHandler,
                 params: MutMap[Quantity], values: MutMap[Quantity]):
        self.__handler = handler
        self.__model_name = "N/A"

        self.__params = params  # reference to dicts in handler
        self.__values = values  # not a copy by design
        self.__free: Map[Quantity] = {}

        self.__provided: set[str] = set()

    def __getitem__(self, name: str) -> Quantity:
        return self.__handler[name]

    def __len__(self) -> int:
        return len(self.__handler)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__handler)

    def set_name(self, name: str):
        """Set the name of the model for better error diagnostics"""
        self.__model_name = name

    def provide(self, **kwargs: Quantity) -> None:
        """Connect a parameter from parent context to child parameter."""
        for name, quantity in kwargs.items():
            self.__assure_ok(name, quantity)
            self.__params[name] = quantity
            self.__provided.add(name)

    def update(self, name: str, value: float, unit: str) -> None:
        """Provide new default value for child parameter"""
        quantity = Quantity(value, unit)
        self.__assure_ok(name, quantity)
        self.__values[name] = quantity

    def __assure_ok(self, name: str, quantity: Quantity) -> None:
        """Check that a new quantity can be processed and is compatible with
        the previously defined slot."""
        model_name = self.__model_name
        if name not in self.__params:
            msg = f"Parameter '{name}' not defined in '{model_name}'"
            raise KeyError(msg)
        if name in self.__provided:
            msg = f"Parameter '{name}' already provided in '{model_name}'"
            raise KeyError(msg)
        # check unit compatibility -> throw exception on demand:
        quantity.to(self.__params[name].units)

    @property
    def free(self) -> Map[Quantity]:
        """Symbols representing the symbols of the parameters that have not
        been provided. These must be used to create an overall function, when
        parameter values are provided from the outside."""
        return dict(self.__free)

    @property
    def values(self) -> Map[Quantity]:
        """Symbols representing the values of the parameters that have not
        been provided. These must be used to call the overall function."""
        return dict(self.__values)

    def finalise(self) -> None:
        """Make sure there are values for all non-provided parameters with
        no value. Remove values of provided parameters"""

        # are all required parameters connected?
        provided = self.__provided
        params = self.__params
        values = set(self.__values.keys())
        all_param = set(params.keys())
        missing = all_param - (provided | values)
        if missing:
            lst = ", ".join(map(lambda m: f"'{m}'", missing))
            name = self.__model_name
            msg = f"Model '{name}' has unresolved parameters: {lst}"
            raise DataFlowError(msg)

        # clean values of provided parameters
        self.__values = {
            name: quantity
            for name, quantity in self.__values.items() if name not in provided
        }

        # create symbols for still free variables (for later overall function)
        self.__free = {
            name: value for name, value in params.items()
            if name in self.__values
        }



"""Unit tests for material related objects"""

from pytest import raises

from simu import (
    InitialState, ThermoFactory, ThermoParameterStore, MaterialDefinition,
    MaterialSpec, SpeciesDefinition, SpeciesDB, flatten_dictionary, Quantity)
from simu.core.model.material import MaterialHandler
from simu.core.thermo.material import MaterialLab
from simu.core.utilities import assert_reproduction
from simu.app import all_contributions, GibbsState
from simu.app.thermo.factories import ExampleThermoFactory


RK_LIQ = "Boston-Mathias-Redlich-Kwong-Liquid"


def test_create_material_definition():
    factory = ExampleThermoFactory()
    species = {"H2O": SpeciesDefinition("H2O")}
    frame = factory.create_frame(species, RK_LIQ)
    store = ThermoParameterStore()
    initial_state = InitialState.from_std(1)
    _ = MaterialDefinition(frame, initial_state, store)


def test_create_material_definition_wrong_init():
    factory = ExampleThermoFactory()
    species = {"H2O": SpeciesDefinition("H2O")}
    frame = factory.create_frame(species, RK_LIQ)
    store = ThermoParameterStore()
    initial_state = InitialState.from_std(2)
    with raises(ValueError) as err:
        _ = MaterialDefinition(frame, initial_state, store)
    assert "Incompatible initial state" in str(err.value)


def test_create_material(material_h2o_rk_liq):
    flat = flatten_dictionary(material_h2o_rk_liq)
    res = {name: f"{value.units:~}" for name, value in flat.items()}
    assert_reproduction(res)


def test_retain_initial_state_material(material_def_h2o_with_param):
    material = material_def_h2o_with_param.create_flow()
    store = material_def_h2o_with_param.store
    param = {store.name: store.get_all_values()}
    x = [300, 2e5, 10]
    material.retain_initial_state(x, param)
    ini = material.initial_state
    assert ini.temperature == Quantity(300.0, "K")
    assert ini.pressure == Quantity(2, "bar")
    assert ini.mol_vector == Quantity([10], "mol")


def test_species_db():
    _ = SpeciesDB({"Water": "H2O", "Ethanol": "C2H5OH"})


def test_material_lab():
    species= SpeciesDB({"Water": "H2O", "Ethanol": "C2H5OH"})
    store = ThermoParameterStore()
    factory = ThermoFactory()
    factory.register_state_definition(GibbsState)
    factory.register(*all_contributions)
    lab = MaterialLab(factory, species, store)

    structure = {
        "state": "GibbsState",
        "contributions": [
            "H0S0ReferenceState",
            "LinearHeatCapacity",
            "StandardState",
            "IdealMix",
            "GibbsIdealGas"]
    }
    initial_state = InitialState.from_std(2)
    definition = lab.define_material(species, initial_state, structure)
    material = definition.create_flow()
    assert_reproduction(list(material.keys()))


def test_handler_def_port():
    handler = MaterialHandler()
    handler.define_port("inlet")


def test_handler_create_proxy():
    handler = MaterialHandler()
    handler.define_port("inlet")
    proxy = handler.create_proxy()
    assert "inlet" in proxy.free_ports()


def test_handler_connect_port(material_h2o_rk_liq):
    handler = MaterialHandler()
    handler.define_port("inlet")
    proxy = handler.create_proxy()
    proxy.connect("inlet", material_h2o_rk_liq)


def test_handler_connect_no_port(material_h2o_rk_liq):
    handler = MaterialHandler()
    proxy = handler.create_proxy()
    with raises(KeyError):
        proxy.connect("inlet", material_h2o_rk_liq)


def test_handler_connect_incompatible_port(material_h2o_rk_liq):
    handler = MaterialHandler()
    handler.define_port("inlet", MaterialSpec(flow=False))
    proxy = handler.create_proxy()
    with raises(ValueError):
        proxy.connect("inlet", material_h2o_rk_liq)


def test_handler_create_material(material_h2o_rk_liq):
    handler = MaterialHandler()
    handler.create_flow("inlet", material_h2o_rk_liq.definition)


def test_handler_create_material_despite_port(material_h2o_rk_liq):
    handler = MaterialHandler()
    handler.define_port("inlet")
    with raises(KeyError):
        handler.create_flow("inlet", material_h2o_rk_liq.definition)



"""module containing classes to obtain thermodynamic parameters from various
sources."""

from typing import Iterable, Collection
from abc import ABC, abstractmethod

from pint import DimensionalityError

from simu.core.utilities import (
    Quantity, parse_quantities_in_struct, SymbolQuantity)
from simu.core.utilities.types import NestedMap, NestedMutMap, MutMap

_RT = tuple[Quantity | NestedMap[Quantity],
            str | NestedMap[str],
            str | NestedMap[str]]


class AbstractThermoSource(ABC):
    """Any source of thermodynamic parameters is to return a Quantity object
    if the path describes an available property.

    Implementations of this class can be added to the
    :class:`ThermoParameterStore` to represent a source of thermodynamic
    parameters. It is advisable to arrange source objects by
    *bibliographic* sources.

    .. note::

        We would have loved to download large databases of thermodynamic
        properties and to publish them with this software. Unfortunate for this
        particular aspect, there are terms like *copyright* and *license* that
        prohibit this.

        While it is completely legal to obtain thermodynamic parameters for free
        from many sources for the sake of working with them, re-publishing
        compiled databases under the LGPL license is another thing.
    """
    @abstractmethod
    def __getitem__(self, path: Iterable[str]) -> Quantity:
        """
        This operator is used to extract a value from the source, addressed by
        a sequence of keys to navigate into the nested data structure:

        >>> struct = {"T": {"H2O": "100 K", "NH3": "200 K"},
        ...           "p": {"H2O": "10 bar", "NH3": "20 atm"}}
        >>> source = StringDictThermoSource(struct)
        >>> print(source["T", "H2O"])
        100 kelvin

        A ``KeyError`` is raised if either a key does not exist or if the final
        key still has sub-keys.
        """
        pass


class NestedDictThermoSource(AbstractThermoSource):
    """A source of thermodynamic parameters defined by a nested dictionary
    of Quantities"""
    def __init__(self, data: NestedMap[Quantity]):
        self.__data = data

    def __getitem__(self, path: Iterable[str]) -> Quantity:
        result = self.__data
        for key in path:
            result = result[key]
        try:
            _ = result.magnitude, result.units  # is it a Quantity?
        except AttributeError:
            raise KeyError("Key doesn't point to Quantity") from None
        return result


class StringDictThermoSource(NestedDictThermoSource):
    """Source of thermodynamic parameters defined by a nested dictionary.
    The leaf entries of the nested dictionary are strings representing
    the quantities, such as "120 K" or "32.1 kJ/(mol*K)"."""
    def __init__(self, data: NestedMap[str]):
        super().__init__(parse_quantities_in_struct(data))


class ThermoParameterStore:
    """This class connects both to the thermodynamic model instances by
    providing the parameters, and to the model's numerical interface by
    providing the symbols and values of all used parameters, allowing those
    to be altered dynamically and optimised on.

    The important aspect here is that a process model can have hundreds of
    thermodynamic model instances, and while they do not need to, most of them
    usually share the same parameters.

    Adapted to the ``SiMu`` data flow, the initial queries by the models to
    fetch parameters are of symbolic nature, using :meth:`get_symbols` and
    :meth:`get_all_symbols`. This happens during the model definition and
    assures that multiple model instances that ask for the same parameter
    receive the same symbol, such that parameter optimisation tasks can be
    accomplished easily.

    In the (numerical) evaluation phase, the store is asked for the actual
    values via :meth:`get_all_values`. In most cases these are used as constants
    in the calculations, but parameter estimation is possible by keeping
    selected parameters as free variables.
    """

    name: str

    def __init__(self):
        self.__provided_parameters: NestedMutMap[Quantity] = {}
        self.__sources: MutMap[AbstractThermoSource] = {}
        self.name = "default"

    def get_symbols(self, parameter_struct: NestedMap[str]) \
            -> NestedMap[Quantity]:
        """Query the nested structure of parameter symbols from the store.
        The ``parameter_struct`` parameter is a nested dictionary with leaf
        entries being the unit of measurement of the thermodynamic parameters
        defined by the path of keys to address them.

        The method returns a dictionary of the same structure, but with
        symbolic quantities as leaf values.

        With multiple calls to this method with varying parameter structs,
        previously defined symbols will be reused, and a
        ``DimensionalityError`` is raised if such previously defined symbol
        is incompatible with respect to the physical dimension.
        """

        def prepare(name: str, key: str, query: NestedMap[str],
                    stored: NestedMutMap[Quantity]):
            """Helper function to recursively retrieve and define symbols"""
            name = f"{name}.{key}" if name else key
            try:
                items = query.items()
            except AttributeError:
                if key in stored:
                    # compare unit for compatibility
                    try:
                        stored[key].to(query)  # convert unit, see if it works
                    except DimensionalityError as err:
                        err.extra_msg = \
                            " - Error fetching previously defined thermo " \
                            f"parameter '{name}'."
                        raise err from None
                else:
                    stored[key] = SymbolQuantity(name, query)
                return stored[key]

            if key not in stored:
                stored[key] = {}

            return {k: prepare(name, k, q, stored[key]) for k, q in items}

        return {k: prepare("", k, s, self.__provided_parameters)
                for k, s in parameter_struct.items()}

    def get_all_symbols(self) -> NestedMap[Quantity]:
        """This method returns all previously prepared symbols, cf.
        :meth:`get_symbols`, as a nested dictionary of symbolic quantities"""
        return self.__provided_parameters

    def get_all_values(self) -> NestedMap[Quantity]:
        """This method seeks in connected data sources for all previously
        prepared symbols.

        A ``KeyError`` is thrown if not all required parameters are available.
        Use :meth:`get_missing_symbols` to get the structure of missing
        parameters.

        The sources are queried sequentially in the reverse sequence of them
        being added, i.e. latest added Sources supersede previous values of
        same parameters.
        """
        found, missing, sources = self.__get_values()
        if missing:
            raise KeyError("Missing parameter values. Use " 
                           "'get_missing_symbols' to find out which")
        return found

    def get_missing_symbols(self) -> NestedMap[str]:
        """This method tries to collect values for all previously prepared
        symbols and returns a nested dictionary of quantities with those not
        found."""
        return self.__get_values()[1]

    def get_sources(self) -> NestedMap[str]:
        """This method returns the name of the data source in which the
        individual values are found"""
        return self.__get_values()[2]

    def add_source(self, name: str, source: AbstractThermoSource):
        """Add a given source of thermodynamic parameters to the store.
        """
        if name in self.__sources:
            raise KeyError(f"Source '{name}' already defined")
        self.__sources[name] = source

    def __get_values(self) -> _RT:
        """Return a tuple of

          a. values found for previously defined symbols, and
          b. symbols for those entries where values are not found.
          c. names of the sources where the values are found
        """
        def get_value(path: Collection[str],
                      qty: Quantity) -> tuple[Quantity, str]:
            """Query the sources for value"""
            for source_name, source in reversed(self.__sources.items()):
                try:
                    result = source[*path]
                except KeyError:
                    continue
                try:
                    result.to(qty.units)
                except DimensionalityError as err:
                    err.extra_msg = \
                        " - Error fetching thermodynamic property " + \
                        ".".join(path)
                    raise err from None
                else:
                    return result, source_name
            else:
                raise KeyError("Parameter not found")

        def extract(path: list[str],
                    struct: Quantity | NestedMutMap[Quantity]) -> _RT:
            """Recursive helper function to extract values"""
            try:
                items = struct.items()
            except AttributeError:  # found a leaf node
                try:
                    value, source = get_value(path, struct)
                    return value, {}, source
                except KeyError:
                    # the space after the unit is important here, for the
                    # string not to evaluate to false in case of dimensionless
                    # parameters!
                    return {}, f"{struct.units:~} ", {}

            # found a sub structure
            found: NestedMap[Quantity] = {}
            missing: NestedMutMap[str] = {}
            source: NestedMap[str] = {}
            for key, value in items:
                found_i, miss, source_i = extract(path + [key], value)
                found[key], source[key] = found_i, source_i
                if miss:
                    missing[key] = miss

            return found, missing, source

        return extract([], self.__provided_parameters)



from simu import Model, NumericHandler, Quantity


class Square(Model):
    """A model of a square"""

    def interface(self):
        self.parameters.define("length", 10, "m")
        self.properties.declare("area", "m^2")

    def define(self):
        self.properties["area"] = self.parameters["length"] ** 2


def main():
    numeric = NumericHandler(Square.top())
    func = numeric.function
    print(func.arg_structure)
    print(func.result_structure)

    args = numeric.arguments
    print(args)

    args[NumericHandler.MODEL_PARAMS]["length"] = Quantity(20, "cm")
    result = func(args)

    print(f"{result[NumericHandler.MODEL_PROPS]['area']:.3fP~}")


if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
"""Test module for basic contributions"""

from pytest import raises

# internal modules
from simu.core.thermo import InitialState, SpeciesDefinition
from simu.app.thermo.contributions import (
    GibbsIdealGas, H0S0ReferenceState, HelmholtzIdealGas, IdealMix,
    LinearHeatCapacity, ConstantGibbsVolume, MolecularWeight, ChargeBalance)
from simu.core.utilities import (
    ParameterDictionary, Quantity, SymbolQuantity, assert_reproduction,
    base_unit)


# auxiliary functions
def sym(name, units):
    """Return a scalar symbol of given name and units"""
    return SymbolQuantity(name, base_unit(units))


def vec(name, size, units):
    """Return a vector symbol of given name, units, and size"""
    return SymbolQuantity(name, base_unit(units), size)


def test_h0s0_reference_state(species_definitions_ab):
    """Test definition of H0S0ReferenceState contribution"""

    res = {"T": sym("T", "K"), "n": vec("n", 2, "mol")}
    bounds = {}
    par = ParameterDictionary()

    cont = H0S0ReferenceState(species_definitions_ab, {})
    cont.define(res)
    to_reproduce = {
        "res": {i: str(res[i])
                for i in "S mu".split()},
        "par_names": list(cont.parameters.keys())
    }
    assert_reproduction(to_reproduce)


def test_linear_heat_capacity(species_definitions_ab):
    """Test definition of LinearHeatCapacity contribution"""
    res = {
        "T": sym("T", "K"),
        "T_ref": sym("T_ref", "K"),
        "n": vec("n", 2, "mol"),
        "S": sym("S_ref", "J/K"),
        "mu": vec("mu_ref", 2, "J/mol")
    }
    bounds = {}
    cont = LinearHeatCapacity(species_definitions_ab, {})
    par = ParameterDictionary()
    cont.define(res)
    result = {i: str(res[i]).split(", ") for i in "S mu".split()}
    assert_reproduction(result)


def test_ideal_mix(species_definitions_ab):
    """Test definition of IdealMix contribution"""
    res = {
        "T": sym("T", "K"),
        "n": vec("n", 2, "mol"),
        "S": sym("S_std", "J/K"),
        "mu": vec("mu_std", 2, "J/mol")
    }
    bounds = {}
    cont = IdealMix(species_definitions_ab, {})
    cont.define(res)
    result = {i: str(res[i]).split(", ") for i in "S mu".split()}
    assert_reproduction(result)


def test_gibbs_ideal_gas(species_definitions_ab):
    """Test definition of GibbsIdealGas contribution"""
    res = {
        "T": sym("T", "K"),
        "p": sym("p", "bar"),
        "n": vec('n', 2, "mol"),
        "p_ref": sym("p_ref", "bar"),
        "S": sym("S_im", "J/K"),
        "mu": vec("mu_im", 2, "J/mol")
    }
    cont = GibbsIdealGas(species_definitions_ab, {})
    cont.define(res)
    result = {i: str(res[i]).split(", ") for i in "S V mu".split()}
    assert_reproduction(result)

    # unit correct?
    assert res["S"].is_compatible_with("J/K")
    assert res["V"].is_compatible_with("m**3")
    assert res["mu"].is_compatible_with("J/mol")


def test_helmholtz_ideal_gas(species_definitions_ab):
    """Test definition of GibbsIdealGas contribution"""
    res = {
        "T": sym("T", "K"),
        "V": sym("V", "m ** 3"),
        "n": vec('n', 2, "mol"),
        "p_ref": sym("p_ref", "bar"),
        "S": sym("S_im", "J/K"),
        "mu": vec("mu_im", 2, "J/mol")
    }
    bounds = {}
    cont = HelmholtzIdealGas(species_definitions_ab, {})
    cont.define(res)
    result = {i: str(res[i]).split(", ") for i in "S p mu".split()}
    assert_reproduction(result)


def test_helmholtz_ideal_gas_initialise(species_definitions_ab):
    """Test initialisation via Helmholtz ideal gas contribution"""
    cont = HelmholtzIdealGas(species_definitions_ab, {})
    # normally, we would need to provide numeric quantities as results,
    #  but these are not used for ideal gas initialisation.
    initial_state = InitialState(temperature=Quantity("25 degC"),
                                 pressure=Quantity("1 bar"),
                                 mol_vector=Quantity([1, 1], "mol"))

    state = cont.initial_state(initial_state, {})
    ref_volume = 2 * 8.31446261815324 * (273.15 + 25) / 1e5
    assert abs(state[1] / ref_volume - 1) < 1e-7


def test_constant_gibbs_volume(species_definitions_ab):
    """Test definition of constant gibbs volume contribution"""
    res = {
        "p": sym("p", "Pa"),
        "p_ref": sym("p_ref", "Pa"),
        "n": vec("n", 2, "mol"),
        "mu": vec("mu_std", 2, "J/mol")
    }
    bounds = {}
    cont = ConstantGibbsVolume(species_definitions_ab, {})
    cont.define(res)
    result = {i: str(res[i]).split(", ") for i in "V mu".split()}
    assert_reproduction(result)


def test_molecular_weight(species_definitions_ab):
    """Test definition of molecular weights"""
    res, bounds = {}, {}
    cont = MolecularWeight(species_definitions_ab, {})
    cont.define(res)
    mw = res["mw"].to("g/mol").magnitude
    assert_reproduction(str(mw))


def test_charge_balance():
    species = {"A": SpeciesDefinition("Na:1+"),
               "B": SpeciesDefinition("Cl:1-")}
    res, bounds = {"n": vec("n", 2, "mol")}, {}
    cont = ChargeBalance(species, {})
    cont.define(res)
    res_str = str(cont.residuals["balance"].value.m)
    assert res_str == "(n_0-n_1)"

def test_charge_balance_only_positive():
    species = {"A": SpeciesDefinition("Na:1+"),
               "B": SpeciesDefinition("K:1+")}
    res, bounds = {"n": vec("n", 2, "mol")}, {}
    cont = ChargeBalance(species, {})
    with raises(ValueError):
        cont.define(res)

def test_no_charge_balance(species_definitions_ab):
    res, bounds = {"n": vec("n", 2, "mol")}, {}
    cont = ChargeBalance(species_definitions_ab, {})
    cont.define(res)
    assert not cont.residuals
